from __future__ import absolute_import
from __future__ import print_function

import math
import functools

import veriloggen.core.vtypes as vtypes
import veriloggen.types.axi as axi
from veriloggen.fsm.fsm import FSM
from veriloggen.optimizer import try_optimize as optimize

from .axim import AXIM
from .axistreamin import AXIStreamIn, AXIStreamInFifo
from .axistreamout import AXIStreamOut, AXIStreamOutFifo

from .fifo import FIFO


class AXIM_for_AXIStreamIn(AXIM, axi.AxiMaster):

    def __init__(self, streamin, name,
                 waddr_id_width=0, wdata_id_width=0, wresp_id_width=0,
                 raddr_id_width=0, rdata_id_width=0,
                 waddr_user_width=2, wdata_user_width=0, wresp_user_width=0,
                 raddr_user_width=2, rdata_user_width=0,
                 waddr_burst_mode=axi.BURST_INCR, raddr_burst_mode=axi.BURST_INCR,
                 waddr_cache_mode=axi.AxCACHE_NONCOHERENT,
                 raddr_cache_mode=axi.AxCACHE_NONCOHERENT,
                 waddr_prot_mode=axi.AxPROT_NONCOHERENT,
                 raddr_prot_mode=axi.AxPROT_NONCOHERENT,
                 waddr_user_mode=axi.AxUSER_NONCOHERENT,
                 wdata_user_mode=axi.xUSER_DEFAULT,
                 raddr_user_mode=axi.AxUSER_NONCOHERENT,
                 noio=False,
                 use_global_base_addr=False,
                 req_fifo_addrwidth=3, fsm_as_module=False):

        if not isinstance(streamin, AXIStreamIn):
            raise TypeError('AXIStreamIn object is required.')

        if not streamin.noio:
            raise ValueError("noio mode of AXIStreamIn is required.")

        m = streamin.m
        clk = streamin.clk
        rst = streamin.rst
        datawidth = streamin.datawidth
        addrwidth = streamin.addrwidth

        axi.AxiMaster.__init__(self, m, name, clk, rst, datawidth, addrwidth,
                               waddr_id_width, wdata_id_width, wresp_id_width,
                               raddr_id_width, rdata_id_width,
                               waddr_user_width, wdata_user_width, wresp_user_width,
                               raddr_user_width, rdata_user_width,
                               waddr_burst_mode, raddr_burst_mode,
                               waddr_cache_mode, raddr_cache_mode,
                               waddr_prot_mode, raddr_prot_mode,
                               waddr_user_mode, wdata_user_mode,
                               raddr_user_mode,
                               noio, req_fifo_addrwidth)

        self.use_global_base_addr = use_global_base_addr
        self.req_fifo_addrwidth = req_fifo_addrwidth
        self.fsm_as_module = fsm_as_module

        self.mutex = None

        # Read
        self.read_start = self.m.Reg('_'.join(['', self.name, 'read_start']),
                                     initval=0)
        self.read_global_addr = self.m.Reg('_'.join(['', self.name, 'read_global_addr']),
                                           self.addrwidth, initval=0)
        self.read_global_size = self.m.Reg('_'.join(['', self.name, 'read_global_size']),
                                           self.addrwidth + 1, initval=0)

        self.read_req_fifo = FIFO(self.m, '_'.join(['', self.name, 'read_req_fifo']),
                                  self.clk, self.rst,
                                  datawidth=self.addrwidth + 1,
                                  addrwidth=self.req_fifo_addrwidth,
                                  sync=False)

        self.read_local_size_fifo = self.m.Wire('_'.join(['', self.name,
                                                          'read_local_size_fifo']),
                                                self.addrwidth + 1)

        self.read_local_size_fifo.assign(self.read_req_fifo.rdata)

        self.read_local_size_buf = self.m.Reg('_'.join(['', self.name,
                                                        'read_local_size_buf']),
                                              self.addrwidth + 1, initval=0)

        self.read_req_busy = self.m.Reg(
            '_'.join(['', self.name, 'read_req_busy']), initval=0)
        self.read_data_busy = self.m.Reg(
            '_'.join(['', self.name, 'read_data_busy']), initval=0)

        self.read_req_idle = self.m.Wire('_'.join(['', self.name, 'read_req_idle']))
        self.read_data_idle = self.m.Wire('_'.join(['', self.name, 'read_data_idle']))
        self.read_idle = self.m.Wire('_'.join(['', self.name, 'read_idle']))

        self.read_req_idle.assign(vtypes.Ands(vtypes.Not(self.read_start),
                                              vtypes.Not(self.read_req_busy)))
        self.read_data_idle.assign(vtypes.Ands(self.read_req_fifo.empty,
                                               vtypes.Not(self.read_data_busy)))
        self.read_idle.assign(vtypes.Ands(self.read_req_idle, self.read_data_idle))

        self.seq(
            self.read_start(0)
        )

        self.read_req_fsm = None
        self.read_data_fsm = None

        if self.use_global_base_addr:
            self.global_base_addr = self.m.Reg('_'.join(['', self.name, 'global_base_addr']),
                                               self.addrwidth, initval=0)
        else:
            self.global_base_addr = None

        # No write
        self.disable_write()

        self.streamin = streamin
        streamout_name = '_'.join((streamin.name, 'out'))
        self.streamout = axi.AxiStreamOut(m, streamout_name, clk, rst,
                                          datawidth=streamin.datawidth,
                                          with_last=streamin.tdata.tlast is not None,
                                          with_strb=streamin.tdata.tstrb is not None,
                                          id_width=streamin.tdata.id_width,
                                          user_width=streamin.tdata.user_width,
                                          dest_width=streamin.tdata.dest_width,
                                          noio=streamin.noio)
        streamin.connect_stream(self.streamout)

    def dma_read(self, fsm, global_addr, size):
        raise NotImplementedError('Blocking dma_read is not supported. Use dma_read_async.')

    def dma_read_async(self, fsm, global_addr, size):

        return self._dma_read(fsm, global_addr, size)

    def dma_write(self, fsm, global_addr, size):
        raise NotImplementedError('Not supported.')

    def dma_write_async(self, fsm, global_addr, size):
        raise NotImplementedError('Not supported.')

    def dma_wait(self, fsm):
        return self.dma_wait_read(fsm)

    def _dma_read(self, fsm, global_addr, size):

        start = vtypes.Ands(fsm.here, self.read_req_idle)

        if not isinstance(self.datawidth, int):
            raise TypeError("axi.datawidth must be int, not '%s'" %
                            str(type(self.datawidth)))

        self._set_read_request(start, global_addr, size)
        self._synthesize_read_req_fsm()
        self._synthesize_read_data_fsm()

        fsm.If(self.read_req_idle).goto_next()

    def _set_read_request(self, start, global_addr, size):

        if self.use_global_base_addr:
            global_addr = global_addr + self.global_base_addr

        self.seq.If(start)(
            self.read_start(1),
            self.read_global_addr(self.mask_addr(global_addr)),
            self.read_global_size(size),
        )

    def _synthesize_read_req_fsm(self):

        if self.read_req_fsm is not None:
            return

        req_fsm = FSM(self.m, '_'.join(['', self.name, 'read_req_fsm']),
                      self.clk, self.rst, as_module=self.fsm_as_module)
        self.read_req_fsm = req_fsm

        cur_global_size = self.m.Reg('_'.join(['', self.name, 'read_cur_global_size']),
                                     self.addrwidth + 1, initval=0)
        cont = self.m.Reg('_'.join(['', self.name, 'read_cont']), initval=0)
        max_burstlen = 2 ** self.burst_size_width

        # Req state 0
        self.seq.If(req_fsm.here, self.read_start)(
            self.read_req_busy(1)
        )
        self.seq.If(self.read_start, self.read_req_fifo.almost_full)(
            self.read_start(1)
        )

        enq_cond = vtypes.Ands(req_fsm.here, self.read_start,
                               vtypes.Not(self.read_req_fifo.almost_full))
        _ = self.read_req_fifo.enq_rtl(self.read_global_size,
                                       cond=enq_cond)

        check_cond = vtypes.Ands(req_fsm.here, vtypes.Ors(self.read_start, cont),
                                 vtypes.Not(self.read_req_fifo.almost_full))
        self._check_4KB_boundary(req_fsm, max_burstlen,
                                 self.read_global_addr, cur_global_size, self.read_global_size,
                                 cond=check_cond)
        req_fsm.If(check_cond).goto_next()

        # Req state 1
        ack = self.read_request(self.read_global_addr, cur_global_size, cond=req_fsm)
        req_fsm.If(ack)(
            cont(1)
        )
        self.seq.If(req_fsm.here, ack)(
            self.read_global_addr.add(optimize(cur_global_size * (self.datawidth // 8)))
        )
        req_fsm.If(ack, self.read_global_size == 0)(
            cont(0)
        )
        self.seq.If(req_fsm.here, ack, self.read_global_size == 0)(
            self.read_req_busy(0)
        )
        req_fsm.If(ack).goto_init()

    def _synthesize_read_data_fsm(self):

        # Data FSM
        if self.read_data_fsm is not None:
            return

        data_fsm = FSM(self.m, '_'.join(['', self.name, 'read_data_fsm']),
                       self.clk, self.rst, as_module=self.fsm_as_module)
        self.read_data_fsm = data_fsm

        # Data state 0
        cond = vtypes.Ands(vtypes.Not(self.read_data_busy),
                           vtypes.Not(self.read_req_fifo.empty))
        self.seq.If(data_fsm.here, cond)(
            self.read_data_busy(1),
            self.read_local_size_buf(self.read_local_size_fifo),
        )
        deq_cond = vtypes.Ands(data_fsm.here, cond)
        _ = self.read_req_fifo.deq_rtl(cond=deq_cond)
        data_fsm.If(cond).goto_next()

        # Data state 1
        rready = vtypes.Ands(vtypes.Ors(self.streamin.tdata.tready,
                                        vtypes.Not(self.streamin.tdata.tvalid)),
                             self.read_local_size_buf > 0)
        _ = self.read_data(cond=rready)

        wlast = self.read_local_size_buf <= 1
        wcond = vtypes.Ands(self.rdata.rvalid, rready)
        # self.streamin.write_data(self.rdata.rdata, wlast, cond=wcond) ### ???
        self.streamout.write_data(self.rdata.rdata, wlast, cond=wcond)

        self.seq.If(data_fsm.here, self.rdata.rvalid, rready)(
            self.read_local_size_buf.dec()
        )

        data_fsm.If(wcond, wlast).goto_init()
        self.seq.If(data_fsm.here, wcond, wlast)(
            self.read_data_busy(0)
        )


class AXIM_for_AXIStreamOut(AXIM, axi.AxiMaster):

    def __init__(self, streamout, name,
                 waddr_id_width=0, wdata_id_width=0, wresp_id_width=0,
                 raddr_id_width=0, rdata_id_width=0,
                 waddr_user_width=2, wdata_user_width=0, wresp_user_width=0,
                 raddr_user_width=2, rdata_user_width=0,
                 waddr_burst_mode=axi.BURST_INCR, raddr_burst_mode=axi.BURST_INCR,
                 waddr_cache_mode=axi.AxCACHE_NONCOHERENT,
                 raddr_cache_mode=axi.AxCACHE_NONCOHERENT,
                 waddr_prot_mode=axi.AxPROT_NONCOHERENT,
                 raddr_prot_mode=axi.AxPROT_NONCOHERENT,
                 waddr_user_mode=axi.AxUSER_NONCOHERENT,
                 wdata_user_mode=axi.xUSER_DEFAULT,
                 raddr_user_mode=axi.AxUSER_NONCOHERENT,
                 noio=False,
                 use_global_base_addr=False,
                 req_fifo_addrwidth=3, fsm_as_module=False):

        if not isinstance(streamout, AXIStreamOut):
            raise TypeError('AXIStreamOut object is required.')

        if not streamout.noio:
            raise ValueError("noio mode of AXIStreamOut is required.")

        m = streamout.m
        clk = streamout.clk
        rst = streamout.rst
        datawidth = streamout.datawidth
        addrwidth = streamout.addrwidth

        axi.AxiMaster.__init__(self, m, name, clk, rst, datawidth, addrwidth,
                               waddr_id_width, wdata_id_width, wresp_id_width,
                               raddr_id_width, rdata_id_width,
                               waddr_user_width, wdata_user_width, wresp_user_width,
                               raddr_user_width, rdata_user_width,
                               waddr_burst_mode, raddr_burst_mode,
                               waddr_cache_mode, raddr_cache_mode,
                               waddr_prot_mode, raddr_prot_mode,
                               waddr_user_mode, wdata_user_mode,
                               raddr_user_mode,
                               noio, req_fifo_addrwidth)

        self.use_global_base_addr = use_global_base_addr
        self.req_fifo_addrwidth = req_fifo_addrwidth
        self.fsm_as_module = fsm_as_module

        self.mutex = None

        # Write
        self.write_start = self.m.Reg('_'.join(['', self.name, 'write_start']),
                                      initval=0)
        self.write_global_addr = self.m.Reg('_'.join(['', self.name, 'write_global_addr']),
                                            self.addrwidth, initval=0)
        self.write_global_size = self.m.Reg('_'.join(['', self.name, 'write_global_size']),
                                            self.addrwidth + 1, initval=0)

        self.write_req_fifo = FIFO(self.m, '_'.join(['', self.name, 'write_req_fifo']),
                                   self.clk, self.rst,
                                   datawidth=self.addrwidth + 1,
                                   addrwidth=self.req_fifo_addrwidth,
                                   sync=False)

        self.write_size_fifo = self.m.Wire('_'.join(['', self.name,
                                                     'write_size_fifo']),
                                           self.addrwidth + 1)
        self.write_size_fifo.assign(self.write_req_fifo.rdata)

        self.write_size_buf = self.m.Reg('_'.join(['', self.name,
                                                   'write_size_buf']),
                                         self.addrwidth + 1, initval=0)

        self.write_req_busy = self.m.Reg(
            '_'.join(['', self.name, 'write_req_busy']), initval=0)
        self.write_data_busy = self.m.Reg(
            '_'.join(['', self.name, 'write_data_busy']), initval=0)

        self.write_req_idle = self.m.Wire('_'.join(['', self.name, 'write_req_idle']))
        self.write_data_idle = self.m.Wire('_'.join(['', self.name, 'write_data_idle']))
        self.write_idle = self.m.Wire('_'.join(['', self.name, 'write_idle']))

        self.write_req_idle.assign(vtypes.Ands(vtypes.Not(self.write_start),
                                               vtypes.Not(self.write_req_busy)))
        self.write_data_idle.assign(vtypes.Ands(self.write_req_fifo.empty,
                                                vtypes.Not(self.write_data_busy)))
        self.write_idle.assign(vtypes.Ands(self.write_req_idle, self.write_data_idle))

        self.seq(
            self.write_start(0)
        )

        self.write_req_fsm = None
        self.write_data_fsm = None

        if self.use_global_base_addr:
            self.global_base_addr = self.m.Reg('_'.join(['', self.name, 'global_base_addr']),
                                               self.addrwidth, initval=0)
        else:
            self.global_base_addr = None

        # No read
        self.disable_read()

        self.streamout = streamout
        streamin_name = '_'.join((streamout.name, 'in'))
        self.streamin = axi.AxiStreamIn(m, streamin_name, clk, rst,
                                        datawidth=streamout.datawidth,
                                        with_last=streamout.tdata.tlast is not None,
                                        with_strb=streamout.tdata.tstrb is not None,
                                        id_width=streamout.tdata.id_width,
                                        user_width=streamout.tdata.user_width,
                                        dest_width=streamout.tdata.dest_width,
                                        noio=streamout.noio)
        streamout.connect_stream(self.streamin)

    def dma_read(self, fsm, global_addr, size):
        raise NotImplementedError('Not supported.')

    def dma_read_async(self, fsm, global_addr, size):
        raise NotImplementedError('Not supported.')

    def dma_write(self, fsm, global_addr, size):
        raise NotImplementedError('Blocking dma_write is not supported. Use dma_write_async.')

    def dma_write_async(self, fsm, global_addr, size):

        return self._dma_write(fsm, global_addr, size)

    def _dma_write(self, fsm, global_addr, size):

        start = vtypes.Ands(fsm.here, self.write_req_idle)

        if not isinstance(self.datawidth, int):
            raise TypeError("axi.datawidth must be int, not '%s'" %
                            str(type(self.datawidth)))

        self._set_write_request(start, global_addr, size)
        self._synthesize_write_req_fsm()
        self._synthesize_write_data_fsm()

        fsm.If(self.write_req_idle).goto_next()

    def _set_write_request(self, start, global_addr, size):

        if self.use_global_base_addr:
            global_addr = global_addr + self.global_base_addr

        self.seq.If(start)(
            self.write_start(1),
            self.write_global_addr(self.mask_addr(global_addr)),
            self.write_global_size(size),
        )

    def _synthesize_write_req_fsm(self):

        if self.write_req_fsm is not None:
            return

        req_fsm = FSM(self.m, '_'.join(['', self.name, 'write_req_fsm']),
                      self.clk, self.rst, as_module=self.fsm_as_module)
        self.write_req_fsm = req_fsm

        cur_global_size = self.m.Reg('_'.join(['', self.name, 'write_cur_global_size']),
                                     self.addrwidth + 1, initval=0)
        cont = self.m.Reg('_'.join(['', self.name, 'write_cont']), initval=0)
        max_burstlen = 2 ** self.burst_size_width

        # Req state 0
        self.seq.If(req_fsm.here, self.write_start)(
            self.write_req_busy(1)
        )
        self.seq.If(self.write_start, self.write_req_fifo.almost_full)(
            self.write_start(1)
        )

        enq_cond = vtypes.Ands(req_fsm.here, self.write_start,
                               vtypes.Not(self.write_req_fifo.almost_full))
        _ = self.write_req_fifo.enq_rtl(self.write_global_size,
                                        cond=enq_cond)

        check_cond = vtypes.Ands(req_fsm.here, vtypes.Ors(self.write_start, cont),
                                 vtypes.Not(self.write_req_fifo.almost_full))
        self._check_4KB_boundary(req_fsm, max_burstlen,
                                 self.write_global_addr, cur_global_size, self.write_global_size,
                                 cond=check_cond)
        req_fsm.If(check_cond).goto_next()

        # Req state 1
        enq_cond = vtypes.Ands(req_fsm.here,
                               vtypes.Not(self.write_req_fifo.almost_full),
                               vtypes.Ors(self.waddr.awready, vtypes.Not(self.waddr.awvalid)),
                               self.write_acceptable())
        _ = self.write_req_fifo.enq_rtl(cur_global_size,
                                        cond=enq_cond)

        req_cond = vtypes.Ands(req_fsm.here,
                               vtypes.Not(self.write_req_fifo.almost_full),
                               self.write_acceptable())
        ack = self.write_request(self.write_global_addr, cur_global_size, cond=req_cond)
        req_fsm.If(enq_cond)(
            cont(1)
        )
        self.seq.If(req_fsm.here, enq_cond)(
            self.write_global_addr.add(optimize(cur_global_size * (self.datawidth // 8)))
        )
        req_fsm.If(enq_cond, self.write_global_size == 0)(
            cont(0)
        )
        self.seq.If(req_fsm.here, enq_cond, self.write_global_size == 0)(
            self.write_req_busy(0)
        )
        req_fsm.If(enq_cond).goto_init()

    def _synthesize_write_data_fsm(self):

        # Data FSM
        if self.write_data_fsm is not None:
            return

        data_fsm = FSM(self.m, '_'.join(['', self.name, 'write_data_fsm']),
                       self.clk, self.rst, as_module=self.fsm_as_module)
        self.write_data_fsm = data_fsm

        # Data state 0
        cond = vtypes.Ands(vtypes.Not(self.write_data_busy),
                           vtypes.Not(self.write_req_fifo.empty))
        count = self.m.TmpReg(self.addrwidth + 1, initval=0,
                              prefix='_'.join(['', self.name, 'write_count']))
        self.seq.If(data_fsm.here, cond)(
            self.write_data_busy(1),
            self.write_size_buf(0),
            count(self.write_size_fifo)
        )
        deq_cond = vtypes.Ands(data_fsm.here, cond)
        _ = self.write_req_fifo.deq_rtl(cond=deq_cond)
        data_fsm.If(cond).goto_next()

        # Data state 1
        cond = vtypes.Ands(vtypes.Not(self.write_req_fifo.empty),
                           self.write_size_buf == 0)
        self.seq.If(data_fsm.here, cond)(
            self.write_size_buf(self.write_size_fifo),
        )
        deq_cond = vtypes.Ands(data_fsm.here, cond)
        _ = self.write_req_fifo.deq_rtl(cond=deq_cond)

        rready = vtypes.Ands(vtypes.Ors(self.wdata.wready, vtypes.Not(self.wdata.wvalid)),
                             self.write_size_buf > 0)
        _ = self.streamin.read_data(cond=rready)

        wlast = self.write_size_buf <= 1
        wcond = vtypes.Ands(self.streamin.tdata.tvalid, rready)
        _ = self.write_data(self.streamin.tdata.tdata, wlast, cond=wcond)

        self.seq.If(data_fsm.here, self.streamin.tdata.tvalid, rready)(
            self.write_size_buf.dec(),
            count.dec()
        )

        data_fsm.If(wcond, count == 1).goto_init()
        self.seq.If(data_fsm.here, wcond, count == 1)(
            self.write_data_busy(0)
        )



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

from veriloggen import *


def mkLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', width)
    count = m.Reg('count', 32)

    m.Always(Posedge(clk))(
        If(rst)(
            count(0)
        ).Else(
            If(count == 1023)(
                count(0)
            ).Else(
                count(count + 1)
            )
        ))

    m.Always(Posedge(clk))(
        If(rst)(
            led(0)
        ).Else(
            If(count == 1024 - 1)(
                led(led + 1)
            )
        ))

    return m


def mkTest():
    m = Module('test')
    width = m.Parameter('WIDTH', 8)
    clk = m.Reg('CLK')
    rst = m.Reg('RST')
    led = m.Wire('LED', width)

    uut = m.Instance(mkLed(), 'uut',
                     params=(('WIDTH', width),),
                     ports=(('CLK', clk), ('RST', rst), ('LED', led)))

    m.Initial(
        Systask('dumpfile', 'initial.vcd'),
        Systask('dumpvars', 0, uut)
    )

    m.Initial(
        clk(0),
        Forever(clk(Not(clk), ldelay=5))  # forever #5 CLK = ~CLK;
    )

    m.Initial(
        rst(0),
        Delay(100),
        rst(1),
        Delay(100),
        rst(0),
        Delay(1000),
        Systask('finish'),
    )

    return m


if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes
from veriloggen.seq.seq import TmpSeq


class SkidBuffer(vtypes.VeriloggenNode):

    def __init__(self, m, clk, rst,
                 s_valid, m_ready, *s_values, prefix=None):

        if prefix is None:
            prefix = 'skidbuffer'

        if not s_values:
            raise ValueError("'s_values' must not be empty.")

        self.m = m
        self.clk = clk
        self.rst = rst
        self.s_values = s_values
        self.s_valid = s_valid
        self.m_ready = m_ready

        width = 0
        _s_values = []
        for s_value in s_values:
            v = m.TmpWireLike(s_value, prefix=prefix + '_s_value')
            v.assign(s_value)
            _s_values.append(v)
            w = s_value.get_width()
            if w is None:
                w = 1
            width += w

        _s_data = m.TmpWire(width, prefix=prefix + '_s_data')
        _s_data.assign(vtypes.Cat(*_s_values))

        _s_valid = m.TmpWire(prefix=prefix + '_s_valid')
        _s_valid.assign(s_valid)
        _m_ready = m.TmpWire(prefix=prefix + '_m_ready')
        _m_ready.assign(m_ready)

        (m_data, m_valid, s_ready) = make_skidbuffer(m, clk, rst,
                                                     _s_data, _s_valid, _m_ready,
                                                     prefix=prefix)

        self.m_values = []
        msb = width - 1
        for s_value in s_values:
            v = m.TmpWireLike(s_value, prefix=prefix + '_m_value')
            w = v.get_width()
            if w is None:
                w = 1
            lsb = msb - w + 1
            v.assign(vtypes.Slice(m_data, msb, lsb))
            self.m_values.append(v)
            msb -= w

        self.m_valid = m_valid
        self.s_ready = s_ready

    @property
    def values(self):
        return self.m_values

    @property
    def valid(self):
        return self.m_valid

    @property
    def ready(self):
        return self.s_ready

    def __getitem__(self, index):
        return self.m_values[index]


def make_skidbuffer(m, clk, rst,
                    s_data, s_valid, m_ready,
                    prefix=None):

    if prefix is None:
        if hasattr(s_data, 'name'):
            prefix = '_'.join([s_data.name, 'skidbuffer'])
        else:
            prefix = 'skidbuffer'

    m_data = m.TmpRegLike(s_data, prefix=prefix + '_data', initval=0)
    m_valid = m.TmpReg(prefix=prefix + '_valid', initval=0)

    s_ready = m.TmpWire(prefix=prefix + '_ready')

    tmp_data = m.TmpRegLike(s_data, prefix=prefix + '_tmp_data', initval=0)
    tmp_valid = m.TmpReg(prefix=prefix + '_tmp_valid', initval=0)

    next_data = m.TmpWireLike(s_data, prefix=prefix + '_next_data')
    next_valid = m.TmpWire(prefix=prefix + '_next_valid')

    s_ready.assign(vtypes.Not(tmp_valid))

    next_data.assign(vtypes.Mux(tmp_valid, tmp_data, s_data))
    next_valid.assign(vtypes.Ors(tmp_valid, s_valid))

    seq = TmpSeq(m, clk, rst, prefix=prefix + '_seq')

    seq.If(vtypes.Ors(m_ready, vtypes.Not(m_valid)))(
        m_data(next_data),
        m_valid(next_valid),
    )
    seq.If(vtypes.Not(tmp_valid), m_valid, vtypes.Not(m_ready))(
        tmp_data(s_data),
        tmp_valid(s_valid),
    )
    seq.If(tmp_valid, m_ready)(
        tmp_valid(0)
    )

    return m_data, m_valid, s_ready



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi


def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    datawidth = 32
    addrwidth = 10

    myaxi = vthread.AXIM(m, 'myaxi', clk, rst, datawidth)
    # If RAM is simultaneously accesseed with DMA, numports must be 2 or more.
    myram = vthread.RAM(m, 'myram', clk, rst, datawidth, addrwidth, numports=2)

    saxi = vthread.AXISLiteRegister(m, 'saxi', clk, rst, datawidth)

    all_ok = m.TmpReg(initval=0)

    def blink(size):
        # wait start
        saxi.wait_flag(0, value=1, resetvalue=0)
        # reset done
        saxi.write(1, 0)

        all_ok.value = True

        for i in range(4):
            print('# iter %d start' % i)
            # Test for 4KB boundary check
            offset = i * 1024 * 16 + (myaxi.boundary_size - 4)
            body(size, offset)
            print('# iter %d end' % i)

        if all_ok:
            print('# verify (local): PASSED')
        else:
            print('# verify (local): FAILED')

        # result
        saxi.write(2, all_ok)

        # done
        saxi.write_flag(1, 1, resetvalue=0)

    def body(size, offset):
        # write
        for i in range(size):
            wdata = i + 100
            myram.write(i, wdata)

        w_laddr = 0
        w_gaddr = offset
        # If RAM is simultaneously accesseed with DMA, different port must be
        # used.
        myaxi.dma_write_async(myram, w_laddr, w_gaddr, size, port=1)
        print('dma_write_async: [%d] -> [%d]' % (w_laddr, w_gaddr))

        # write
        for i in range(size):
            wdata = i + 1000
            myram.write(i + size, wdata)

        myaxi.dma_wait_write()
        print('dma_wait_write : [%d] -> [%d]' % (w_laddr, w_gaddr))

        w_laddr = size
        w_gaddr = (size + size) * 4 + offset
        myaxi.dma_write_async(myram, w_laddr, w_gaddr, size, port=1)
        print('dma_write_async: [%d] -> [%d]' % (w_laddr, w_gaddr))

        # read
        r_laddr = 0
        r_gaddr = offset
        myaxi.dma_read_async(myram, r_laddr, r_gaddr, size, port=0)
        print('dma_read_async : [%d] <- [%d]' % (r_laddr, r_gaddr))

        for sleep in range(size):
            pass

        myaxi.dma_wait_write()
        print('dma_wait_write : [%d] -> [%d]' % (w_laddr, w_gaddr))

        myaxi.dma_wait_read()
        print('dma_wait_read  : [%d] <- [%d]' % (r_laddr, r_gaddr))

        for i in range(size):
            rdata = myram.read(i)
            if vthread.verilog.NotEql(rdata, i + 100):
                print('rdata[%d] = %d' % (i, rdata))
                all_ok.value = False

        # read
        r_laddr = 0
        r_gaddr = (size + size) * 4 + offset
        myaxi.dma_read(myram, r_laddr, r_gaddr, size, port=0)
        print('dma_read       : [%d] <- [%d]' % (r_laddr, r_gaddr))

        for sleep in range(size):
            pass

        for i in range(size):
            rdata = myram.read(i)
            if vthread.verilog.NotEql(rdata, i + 1000):
                print('rdata[%d] = %d' % (i, rdata))
                all_ok.value = False

    th = vthread.Thread(m, 'th_blink', clk, rst, blink)
    fsm = th.start(32)

    return m


def mkTestTop(memimg_name=None):
    m = Module('test_top')

    # target instance
    led = mkLed()

    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']

    memory = axi.AxiMultiportMemoryModel(m, 'memory', clk, rst, numports=2,
                                         memimg_name=memimg_name)

    r_ports, w_ports = axi.split_read_write(m, ports, 'myaxi')

    memory.connect(0, r_ports, 'r_myaxi')
    memory.connect(1, w_ports, 'w_myaxi')

    # AXI-Slave controller
    _saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(ports, 'saxi')

    def ctrl():
        for i in range(100):
            pass

        for i in range(16):
            # byte addressing
            v = memory.read(i * 4)
            print('read:  mem[%d] -> %x' % (i, v))
            v = v + 1024
            # byte addressing
            memory.write(i * 4, v)
            print('write: mem[%d] <- %x' % (i, v))

        awaddr = 0
        _saxi.write(awaddr, 1)

        araddr = 4
        v = _saxi.read(araddr)
        while v == 0:
            v = _saxi.read(araddr)

        araddr = 8
        v = _saxi.read(araddr)
        if v:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    # vcd_name = os.path.splitext(os.path.basename(__file__))[0] + '.vcd'
    # simulation.setup_waveform(m, uut, dumpfile=vcd_name)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000000),
        Systask('finish'),
    )

    return m


def mkTest(memimg_name=None):

    m = Module('test')

    # target instance
    test = mkTestTop(memimg_name)
    params = m.copy_params(test)
    ports = m.copy_sim_ports(test)
    uut = m.Instance(test, 'uut',
                     params=m.connect_params(test),
                     ports=m.connect_ports(test))
    # vcd_name = os.path.splitext(os.path.basename(__file__))[0] + '.vcd'
    # simulation.setup_waveform(m, uut, dumpfile=vcd_name)

    return m


def run(filename='tmp.v', simtype='iverilog', outputfile=None):

    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    test = mkTest(memimg_name=memimg_name)

    if filename is not None:
        test.to_verilog(filename)

    sim = simulation.Simulator(test, sim=simtype)
    rslt = sim.run(outputfile=outputfile)

    return rslt


if __name__ == '__main__':
    rslt = run(filename='tmp.v')
    print(rslt)




# coding: utf-8

# In[1]:


from pynq import Overlay, allocate
import numpy as np


# In[2]:


bitfile = "thread_memcpy_ipxact.bit"
overlay = Overlay(bitfile)
overlay.ip_dict.keys()


# In[3]:


memcpy = overlay.memcpy_0


# In[4]:


num_words = 1024 * 128

a = allocate(shape=(num_words,), dtype=np.int32)
b = allocate(shape=(num_words,), dtype=np.int32)

a_addr = a.physical_address
b_addr = b.physical_address


# In[5]:


a[:] = np.random.randint(-20, 20, (num_words,), dtype=np.int32)
b[:] = np.zeros((num_words,), dtype=np.int32)
print(b[-16:])


# In[6]:


copy_bytes = a.nbytes

# copy_bytes, a_offset, b_offset
memcpy.write(2 * 4, copy_bytes)
memcpy.write(3 * 4, a_addr)
memcpy.write(4 * 4, b_addr)

# start
memcpy.write(0 * 4, 1)

# busy wait
while True:
    busy = memcpy.read(1 * 4)
    if not busy:
        break


# In[7]:


print(b[-16:])


# In[8]:


expected = a
print(expected[-16:])


# In[9]:


diff_sum = np.sum(expected - b)
print(diff_sum)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

from veriloggen import *


def mkLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    a = m.Input('A', 32)
    b = m.Output('B', width)

    b.assign(a[32 - width:32])

    return m


if __name__ == '__main__':
    led = mkLed()
    verilog = led.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *


def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    valid = m.OutputReg('valid', initval=0)
    counter = m.Reg('counter', 8, initval=0)

    fsm = FSM(m, 'fsm', clk, rst)

    s0 = fsm.init
    s1 = fsm.State(s0).If(valid).goto_next()
    s2 = fsm.State(s1).If(valid).goto_next()
    s3 = fsm.State(s2).If(counter[0] == 0).goto(fsm.next)
    s1_ = fsm.State(s2).Elif(counter[1] == 1).goto(s1)
    s2_ = fsm.State(s2).Else.goto(s2)
    s0_ = fsm.State(s3).goto(s0)

    fsm.Always.If(counter <= 2 ** 8 - 1)(
        counter.inc()
    ).Else(
        counter(0)
    )

    fsm.State(s0).If(counter == 10)(
        valid(0)
    ).Else(
        valid(1)
    )

    fsm.State(s1).If(counter == 20)(
        valid(0)
    ).Else(
        valid(1)
    )

    fsm.State(s2).If(counter == 30)(
        valid(0)
    ).Else(
        valid(1)
    )

    fsm.State(s0_).If(counter == 40)(
        valid(0)
    ).Else(
        valid(1)
    )

    return m


def mkTest():
    m = Module('test')
    clk = m.Reg('CLK')
    rst = m.Reg('RST')
    valid = m.Wire('valid')

    uut = m.Instance(mkLed(), 'uut',
                     ports=(('CLK', clk), ('RST', rst), ('valid', valid)))

    # vcd_name = os.path.splitext(os.path.basename(__file__))[0] + '.vcd'
    # simulation.setup_waveform(m, uut, dumpfile=vcd_name)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, period=100)

    init.add(
        Delay(1000),
        Systask('finish'),
    )

    return m


if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi


def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    datawidth = 32
    saxi = vthread.AXISLiteRegister(m, 'saxi', clk, rst, datawidth, length=4)

    def blink():

        while True:
            saxi.wait_flag(0, value=1, resetvalue=0)
            saxi.write(1, 1)  # set busy
            size = saxi.read(2)

            sum = 0
            for i in range(size):
                sum += i

            saxi.write(3, sum)
            saxi.write(1, 0)  # unset busy

        vthread.finish()

    th = vthread.Thread(m, 'th_blink', clk, rst, blink)
    fsm = th.start()

    return m


def mkTest(memimg_name=None):
    m = Module('test')

    # target instance
    led = mkLed()

    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']

    # memory = axi.AxiMemoryModel(m, 'memory', clk, rst, memimg_name=memimg_name)
    # memory.connect(ports, 'myaxi')

    # AXI-Slave controller
    _saxi = vthread.AXIMLiteVerify(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(ports, 'saxi')

    k = 100
    expected_sum = 0
    for i in range(k):
        expected_sum += i

    def ctrl():
        for i in range(100):
            pass

        # size
        awaddr = 8
        _saxi.write_delayed(awaddr, k, 10)

        # start
        awaddr = 0
        _saxi.write_delayed(awaddr, 1, 10)

        for _ in range(10):
            pass

        # busy check
        araddr = 4
        v = _saxi.read_delayed(araddr, 10)
        while v != 0:
            v = _saxi.read_delayed(araddr, 10)

        # result
        araddr = 12
        v = _saxi.read_delayed(araddr, 10)

        print('result = %d, expected = %d' % (v, expected_sum))

        if v == expected_sum:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

        vthread.finish()

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    # vcd_name = os.path.splitext(os.path.basename(__file__))[0] + '.vcd'
    # simulation.setup_waveform(m, uut, dumpfile=vcd_name)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000000),
        Systask('finish'),
    )

    return m


def run(filename='tmp.v', simtype='iverilog', outputfile=None):

    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    test = mkTest(memimg_name=memimg_name)

    if filename is not None:
        test.to_verilog(filename)

    sim = simulation.Simulator(test, sim=simtype)
    rslt = sim.run(outputfile=outputfile)

    return rslt


if __name__ == '__main__':
    rslt = run(filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function

import os
import veriloggen
import thread_stream_axi_stream


expected_verilog = """
module blinkled
(
  input CLK,
  input RST,
  input [32-1:0] axi_a_tdata,
  input axi_a_tvalid,
  output axi_a_tready,
  input axi_a_tlast,
  input [32-1:0] axi_b_tdata,
  input axi_b_tvalid,
  output axi_b_tready,
  input axi_b_tlast,
  output reg [32-1:0] axi_c_tdata,
  output reg axi_c_tvalid,
  input axi_c_tready,
  output reg axi_c_tlast,
  input [32-1:0] saxi_awaddr,
  input [4-1:0] saxi_awcache,
  input [3-1:0] saxi_awprot,
  input saxi_awvalid,
  output saxi_awready,
  input [32-1:0] saxi_wdata,
  input [4-1:0] saxi_wstrb,
  input saxi_wvalid,
  output saxi_wready,
  output [2-1:0] saxi_bresp,
  output reg saxi_bvalid,
  input saxi_bready,
  input [32-1:0] saxi_araddr,
  input [4-1:0] saxi_arcache,
  input [3-1:0] saxi_arprot,
  input saxi_arvalid,
  output saxi_arready,
  output reg [32-1:0] saxi_rdata,
  output [2-1:0] saxi_rresp,
  output reg saxi_rvalid,
  input saxi_rready
);

  wire _axi_a_read_req_fifo_enq;
  wire [105-1:0] _axi_a_read_req_fifo_wdata;
  wire _axi_a_read_req_fifo_full;
  wire _axi_a_read_req_fifo_almost_full;
  wire _axi_a_read_req_fifo_deq;
  wire [105-1:0] _axi_a_read_req_fifo_rdata;
  wire _axi_a_read_req_fifo_empty;
  wire _axi_a_read_req_fifo_almost_empty;

  _axi_a_read_req_fifo
  inst__axi_a_read_req_fifo
  (
    .CLK(CLK),
    .RST(RST),
    ._axi_a_read_req_fifo_enq(_axi_a_read_req_fifo_enq),
    ._axi_a_read_req_fifo_wdata(_axi_a_read_req_fifo_wdata),
    ._axi_a_read_req_fifo_full(_axi_a_read_req_fifo_full),
    ._axi_a_read_req_fifo_almost_full(_axi_a_read_req_fifo_almost_full),
    ._axi_a_read_req_fifo_deq(_axi_a_read_req_fifo_deq),
    ._axi_a_read_req_fifo_rdata(_axi_a_read_req_fifo_rdata),
    ._axi_a_read_req_fifo_empty(_axi_a_read_req_fifo_empty),
    ._axi_a_read_req_fifo_almost_empty(_axi_a_read_req_fifo_almost_empty)
  );

  reg [4-1:0] count__axi_a_read_req_fifo;
  wire [8-1:0] _axi_a_read_op_sel_fifo;
  wire [32-1:0] _axi_a_read_local_addr_fifo;
  wire [32-1:0] _axi_a_read_local_stride_fifo;
  wire [33-1:0] _axi_a_read_local_size_fifo;
  wire [8-1:0] unpack_read_req_op_sel_0;
  wire [32-1:0] unpack_read_req_local_addr_1;
  wire [32-1:0] unpack_read_req_local_stride_2;
  wire [33-1:0] unpack_read_req_local_size_3;
  assign unpack_read_req_op_sel_0 = _axi_a_read_req_fifo_rdata[104:97];
  assign unpack_read_req_local_addr_1 = _axi_a_read_req_fifo_rdata[96:65];
  assign unpack_read_req_local_stride_2 = _axi_a_read_req_fifo_rdata[64:33];
  assign unpack_read_req_local_size_3 = _axi_a_read_req_fifo_rdata[32:0];
  assign _axi_a_read_op_sel_fifo = unpack_read_req_op_sel_0;
  assign _axi_a_read_local_addr_fifo = unpack_read_req_local_addr_1;
  assign _axi_a_read_local_stride_fifo = unpack_read_req_local_stride_2;
  assign _axi_a_read_local_size_fifo = unpack_read_req_local_size_3;
  reg [8-1:0] _axi_a_read_op_sel_buf;
  reg [32-1:0] _axi_a_read_local_addr_buf;
  reg [32-1:0] _axi_a_read_local_stride_buf;
  reg [33-1:0] _axi_a_read_local_size_buf;
  reg _axi_a_read_data_busy;
  wire _axi_a_read_data_idle;
  wire _axi_a_read_idle;
  assign _axi_a_read_data_idle = _axi_a_read_req_fifo_empty && !_axi_a_read_data_busy;
  assign _axi_a_read_idle = _axi_a_read_data_idle;
  wire _axi_b_read_req_fifo_enq;
  wire [105-1:0] _axi_b_read_req_fifo_wdata;
  wire _axi_b_read_req_fifo_full;
  wire _axi_b_read_req_fifo_almost_full;
  wire _axi_b_read_req_fifo_deq;
  wire [105-1:0] _axi_b_read_req_fifo_rdata;
  wire _axi_b_read_req_fifo_empty;
  wire _axi_b_read_req_fifo_almost_empty;

  _axi_b_read_req_fifo
  inst__axi_b_read_req_fifo
  (
    .CLK(CLK),
    .RST(RST),
    ._axi_b_read_req_fifo_enq(_axi_b_read_req_fifo_enq),
    ._axi_b_read_req_fifo_wdata(_axi_b_read_req_fifo_wdata),
    ._axi_b_read_req_fifo_full(_axi_b_read_req_fifo_full),
    ._axi_b_read_req_fifo_almost_full(_axi_b_read_req_fifo_almost_full),
    ._axi_b_read_req_fifo_deq(_axi_b_read_req_fifo_deq),
    ._axi_b_read_req_fifo_rdata(_axi_b_read_req_fifo_rdata),
    ._axi_b_read_req_fifo_empty(_axi_b_read_req_fifo_empty),
    ._axi_b_read_req_fifo_almost_empty(_axi_b_read_req_fifo_almost_empty)
  );

  reg [4-1:0] count__axi_b_read_req_fifo;
  wire [8-1:0] _axi_b_read_op_sel_fifo;
  wire [32-1:0] _axi_b_read_local_addr_fifo;
  wire [32-1:0] _axi_b_read_local_stride_fifo;
  wire [33-1:0] _axi_b_read_local_size_fifo;
  wire [8-1:0] unpack_read_req_op_sel_4;
  wire [32-1:0] unpack_read_req_local_addr_5;
  wire [32-1:0] unpack_read_req_local_stride_6;
  wire [33-1:0] unpack_read_req_local_size_7;
  assign unpack_read_req_op_sel_4 = _axi_b_read_req_fifo_rdata[104:97];
  assign unpack_read_req_local_addr_5 = _axi_b_read_req_fifo_rdata[96:65];
  assign unpack_read_req_local_stride_6 = _axi_b_read_req_fifo_rdata[64:33];
  assign unpack_read_req_local_size_7 = _axi_b_read_req_fifo_rdata[32:0];
  assign _axi_b_read_op_sel_fifo = unpack_read_req_op_sel_4;
  assign _axi_b_read_local_addr_fifo = unpack_read_req_local_addr_5;
  assign _axi_b_read_local_stride_fifo = unpack_read_req_local_stride_6;
  assign _axi_b_read_local_size_fifo = unpack_read_req_local_size_7;
  reg [8-1:0] _axi_b_read_op_sel_buf;
  reg [32-1:0] _axi_b_read_local_addr_buf;
  reg [32-1:0] _axi_b_read_local_stride_buf;
  reg [33-1:0] _axi_b_read_local_size_buf;
  reg _axi_b_read_data_busy;
  wire _axi_b_read_data_idle;
  wire _axi_b_read_idle;
  assign _axi_b_read_data_idle = _axi_b_read_req_fifo_empty && !_axi_b_read_data_busy;
  assign _axi_b_read_idle = _axi_b_read_data_idle;
  wire _axi_c_write_req_fifo_enq;
  wire [105-1:0] _axi_c_write_req_fifo_wdata;
  wire _axi_c_write_req_fifo_full;
  wire _axi_c_write_req_fifo_almost_full;
  wire _axi_c_write_req_fifo_deq;
  wire [105-1:0] _axi_c_write_req_fifo_rdata;
  wire _axi_c_write_req_fifo_empty;
  wire _axi_c_write_req_fifo_almost_empty;

  _axi_c_write_req_fifo
  inst__axi_c_write_req_fifo
  (
    .CLK(CLK),
    .RST(RST),
    ._axi_c_write_req_fifo_enq(_axi_c_write_req_fifo_enq),
    ._axi_c_write_req_fifo_wdata(_axi_c_write_req_fifo_wdata),
    ._axi_c_write_req_fifo_full(_axi_c_write_req_fifo_full),
    ._axi_c_write_req_fifo_almost_full(_axi_c_write_req_fifo_almost_full),
    ._axi_c_write_req_fifo_deq(_axi_c_write_req_fifo_deq),
    ._axi_c_write_req_fifo_rdata(_axi_c_write_req_fifo_rdata),
    ._axi_c_write_req_fifo_empty(_axi_c_write_req_fifo_empty),
    ._axi_c_write_req_fifo_almost_empty(_axi_c_write_req_fifo_almost_empty)
  );

  reg [4-1:0] count__axi_c_write_req_fifo;
  wire [8-1:0] _axi_c_write_op_sel_fifo;
  wire [32-1:0] _axi_c_write_local_addr_fifo;
  wire [32-1:0] _axi_c_write_local_stride_fifo;
  wire [33-1:0] _axi_c_write_size_fifo;
  wire [8-1:0] unpack_write_req_op_sel_8;
  wire [32-1:0] unpack_write_req_local_addr_9;
  wire [32-1:0] unpack_write_req_local_stride_10;
  wire [33-1:0] unpack_write_req_local_size_11;
  assign unpack_write_req_op_sel_8 = _axi_c_write_req_fifo_rdata[104:97];
  assign unpack_write_req_local_addr_9 = _axi_c_write_req_fifo_rdata[96:65];
  assign unpack_write_req_local_stride_10 = _axi_c_write_req_fifo_rdata[64:33];
  assign unpack_write_req_local_size_11 = _axi_c_write_req_fifo_rdata[32:0];
  assign _axi_c_write_op_sel_fifo = unpack_write_req_op_sel_8;
  assign _axi_c_write_local_addr_fifo = unpack_write_req_local_addr_9;
  assign _axi_c_write_local_stride_fifo = unpack_write_req_local_stride_10;
  assign _axi_c_write_size_fifo = unpack_write_req_local_size_11;
  reg [8-1:0] _axi_c_write_op_sel_buf;
  reg [32-1:0] _axi_c_write_local_addr_buf;
  reg [32-1:0] _axi_c_write_local_stride_buf;
  reg [33-1:0] _axi_c_write_size_buf;
  reg _axi_c_write_data_busy;
  wire _axi_c_write_data_idle;
  wire _axi_c_write_idle;
  assign _axi_c_write_data_idle = _axi_c_write_req_fifo_empty && !_axi_c_write_data_busy;
  assign _axi_c_write_idle = _axi_c_write_data_idle;
  assign saxi_bresp = 0;
  assign saxi_rresp = 0;
  reg signed [32-1:0] _saxi_register_0;
  reg signed [32-1:0] _saxi_register_1;
  reg signed [32-1:0] _saxi_register_2;
  reg signed [32-1:0] _saxi_register_3;
  reg _saxi_flag_0;
  reg _saxi_flag_1;
  reg _saxi_flag_2;
  reg _saxi_flag_3;
  reg signed [32-1:0] _saxi_resetval_0;
  reg signed [32-1:0] _saxi_resetval_1;
  reg signed [32-1:0] _saxi_resetval_2;
  reg signed [32-1:0] _saxi_resetval_3;
  localparam _saxi_maskwidth = 2;
  localparam _saxi_mask = { _saxi_maskwidth{ 1'd1 } };
  localparam _saxi_shift = 2;
  reg [32-1:0] _saxi_register_fsm;
  localparam _saxi_register_fsm_init = 0;
  reg [32-1:0] addr_12;
  reg writevalid_13;
  reg readvalid_14;
  reg prev_awvalid_15;
  reg prev_arvalid_16;
  assign saxi_awready = (_saxi_register_fsm == 0) && (!writevalid_13 && !readvalid_14 && !saxi_bvalid && prev_awvalid_15);
  assign saxi_arready = (_saxi_register_fsm == 0) && (!readvalid_14 && !writevalid_13 && prev_arvalid_16 && !prev_awvalid_15);
  reg [_saxi_maskwidth-1:0] axis_maskaddr_17;
  wire signed [32-1:0] axislite_rdata_18;
  assign axislite_rdata_18 = (axis_maskaddr_17 == 0)? _saxi_register_0 : 
                             (axis_maskaddr_17 == 1)? _saxi_register_1 : 
                             (axis_maskaddr_17 == 2)? _saxi_register_2 : 
                             (axis_maskaddr_17 == 3)? _saxi_register_3 : 'hx;
  wire axislite_flag_19;
  assign axislite_flag_19 = (axis_maskaddr_17 == 0)? _saxi_flag_0 : 
                            (axis_maskaddr_17 == 1)? _saxi_flag_1 : 
                            (axis_maskaddr_17 == 2)? _saxi_flag_2 : 
                            (axis_maskaddr_17 == 3)? _saxi_flag_3 : 'hx;
  wire signed [32-1:0] axislite_resetval_20;
  assign axislite_resetval_20 = (axis_maskaddr_17 == 0)? _saxi_resetval_0 : 
                                (axis_maskaddr_17 == 1)? _saxi_resetval_1 : 
                                (axis_maskaddr_17 == 2)? _saxi_resetval_2 : 
                                (axis_maskaddr_17 == 3)? _saxi_resetval_3 : 'hx;
  reg _saxi_rdata_cond_0_1;
  assign saxi_wready = _saxi_register_fsm == 3;
  wire [10-1:0] ram_a_0_addr;
  wire [32-1:0] ram_a_0_rdata;
  wire [32-1:0] ram_a_0_wdata;
  wire ram_a_0_wenable;
  wire ram_a_0_enable;
  wire [10-1:0] ram_a_1_addr;
  wire [32-1:0] ram_a_1_rdata;
  wire [32-1:0] ram_a_1_wdata;
  wire ram_a_1_wenable;
  wire ram_a_1_enable;
  assign ram_a_0_wdata = 'hx;
  assign ram_a_0_wenable = 0;

  ram_a
  inst_ram_a
  (
    .CLK(CLK),
    .ram_a_0_addr(ram_a_0_addr),
    .ram_a_0_rdata(ram_a_0_rdata),
    .ram_a_0_wdata(ram_a_0_wdata),
    .ram_a_0_wenable(ram_a_0_wenable),
    .ram_a_0_enable(ram_a_0_enable),
    .ram_a_1_addr(ram_a_1_addr),
    .ram_a_1_rdata(ram_a_1_rdata),
    .ram_a_1_wdata(ram_a_1_wdata),
    .ram_a_1_wenable(ram_a_1_wenable),
    .ram_a_1_enable(ram_a_1_enable)
  );

  wire [10-1:0] ram_b_0_addr;
  wire [32-1:0] ram_b_0_rdata;
  wire [32-1:0] ram_b_0_wdata;
  wire ram_b_0_wenable;
  wire ram_b_0_enable;
  wire [10-1:0] ram_b_1_addr;
  wire [32-1:0] ram_b_1_rdata;
  wire [32-1:0] ram_b_1_wdata;
  wire ram_b_1_wenable;
  wire ram_b_1_enable;
  assign ram_b_0_wdata = 'hx;
  assign ram_b_0_wenable = 0;

  ram_b
  inst_ram_b
  (
    .CLK(CLK),
    .ram_b_0_addr(ram_b_0_addr),
    .ram_b_0_rdata(ram_b_0_rdata),
    .ram_b_0_wdata(ram_b_0_wdata),
    .ram_b_0_wenable(ram_b_0_wenable),
    .ram_b_0_enable(ram_b_0_enable),
    .ram_b_1_addr(ram_b_1_addr),
    .ram_b_1_rdata(ram_b_1_rdata),
    .ram_b_1_wdata(ram_b_1_wdata),
    .ram_b_1_wenable(ram_b_1_wenable),
    .ram_b_1_enable(ram_b_1_enable)
  );

  wire [10-1:0] ram_c_0_addr;
  wire [32-1:0] ram_c_0_rdata;
  wire [32-1:0] ram_c_0_wdata;
  wire ram_c_0_wenable;
  wire ram_c_0_enable;
  wire [10-1:0] ram_c_1_addr;
  wire [32-1:0] ram_c_1_rdata;
  wire [32-1:0] ram_c_1_wdata;
  wire ram_c_1_wenable;
  wire ram_c_1_enable;
  assign ram_c_1_wdata = 'hx;
  assign ram_c_1_wenable = 0;

  ram_c
  inst_ram_c
  (
    .CLK(CLK),
    .ram_c_0_addr(ram_c_0_addr),
    .ram_c_0_rdata(ram_c_0_rdata),
    .ram_c_0_wdata(ram_c_0_wdata),
    .ram_c_0_wenable(ram_c_0_wenable),
    .ram_c_0_enable(ram_c_0_enable),
    .ram_c_1_addr(ram_c_1_addr),
    .ram_c_1_rdata(ram_c_1_rdata),
    .ram_c_1_wdata(ram_c_1_wdata),
    .ram_c_1_wenable(ram_c_1_wenable),
    .ram_c_1_enable(ram_c_1_enable)
  );

  reg _mystream_stream_ivalid;
  wire _mystream_stream_oready;
  wire _mystream_stream_internal_oready;
  assign _mystream_stream_internal_oready = 1;
  assign _mystream_stream_oready = _mystream_stream_internal_oready;
  reg [32-1:0] _mystream_fsm;
  localparam _mystream_fsm_init = 0;
  wire _mystream_run_flag;
  reg _mystream_source_start;
  wire _mystream_source_stop;
  reg _mystream_source_busy;
  wire _mystream_sink_start;
  wire _mystream_sink_stop;
  wire _mystream_sink_busy;
  wire _mystream_busy;
  reg _mystream_busy_reg;
  wire _mystream_is_root;
  assign _mystream_is_root = 1;
  reg _mystream_a_idle;
  reg [33-1:0] _mystream_a_source_count;
  reg [5-1:0] _mystream_a_source_mode;
  reg [16-1:0] _mystream_a_source_generator_id;
  reg [32-1:0] _mystream_a_source_offset;
  reg [33-1:0] _mystream_a_source_size;
  reg [32-1:0] _mystream_a_source_stride;
  reg [32-1:0] _mystream_a_source_offset_buf;
  reg [33-1:0] _mystream_a_source_size_buf;
  reg [32-1:0] _mystream_a_source_stride_buf;
  reg [8-1:0] _mystream_a_source_sel;
  reg [32-1:0] _mystream_a_source_ram_raddr;
  reg _mystream_a_source_ram_renable;
  wire [32-1:0] _mystream_a_source_ram_rdata;
  reg _mystream_a_source_fifo_deq;
  wire [32-1:0] _mystream_a_source_fifo_rdata;
  reg [32-1:0] _mystream_a_source_empty_data;
  reg _mystream_b_idle;
  reg [33-1:0] _mystream_b_source_count;
  reg [5-1:0] _mystream_b_source_mode;
  reg [16-1:0] _mystream_b_source_generator_id;
  reg [32-1:0] _mystream_b_source_offset;
  reg [33-1:0] _mystream_b_source_size;
  reg [32-1:0] _mystream_b_source_stride;
  reg [32-1:0] _mystream_b_source_offset_buf;
  reg [33-1:0] _mystream_b_source_size_buf;
  reg [32-1:0] _mystream_b_source_stride_buf;
  reg [8-1:0] _mystream_b_source_sel;
  reg [32-1:0] _mystream_b_source_ram_raddr;
  reg _mystream_b_source_ram_renable;
  wire [32-1:0] _mystream_b_source_ram_rdata;
  reg _mystream_b_source_fifo_deq;
  wire [32-1:0] _mystream_b_source_fifo_rdata;
  reg [32-1:0] _mystream_b_source_empty_data;
  reg [33-1:0] _mystream_c_sink_count;
  reg [5-1:0] _mystream_c_sink_mode;
  reg [16-1:0] _mystream_c_sink_generator_id;
  reg [32-1:0] _mystream_c_sink_offset;
  reg [33-1:0] _mystream_c_sink_size;
  reg [32-1:0] _mystream_c_sink_stride;
  reg [32-1:0] _mystream_c_sink_offset_buf;
  reg [33-1:0] _mystream_c_sink_size_buf;
  reg [32-1:0] _mystream_c_sink_stride_buf;
  reg [8-1:0] _mystream_c_sink_sel;
  reg [32-1:0] _mystream_c_sink_waddr;
  reg _mystream_c_sink_wenable;
  reg [32-1:0] _mystream_c_sink_wdata;
  reg _mystream_c_sink_fifo_enq;
  reg [32-1:0] _mystream_c_sink_fifo_wdata;
  reg [32-1:0] _mystream_c_sink_immediate;
  reg [32-1:0] th_comp;
  localparam th_comp_init = 0;
  reg signed [32-1:0] _th_comp_size_0;
  reg signed [32-1:0] _th_comp_offset_1;
  wire [8-1:0] pack_read_req_op_sel_21;
  wire [32-1:0] pack_read_req_local_addr_22;
  wire [32-1:0] pack_read_req_local_stride_23;
  wire [33-1:0] pack_read_req_local_size_24;
  assign pack_read_req_op_sel_21 = 1;
  assign pack_read_req_local_addr_22 = _th_comp_offset_1;
  assign pack_read_req_local_stride_23 = 1;
  assign pack_read_req_local_size_24 = _th_comp_size_0;
  wire [105-1:0] pack_read_req_packed_25;
  assign pack_read_req_packed_25 = { pack_read_req_op_sel_21, pack_read_req_local_addr_22, pack_read_req_local_stride_23, pack_read_req_local_size_24 };
  assign _axi_a_read_req_fifo_wdata = ((th_comp == 6) && !_axi_a_read_req_fifo_almost_full)? pack_read_req_packed_25 : 'hx;
  assign _axi_a_read_req_fifo_enq = ((th_comp == 6) && !_axi_a_read_req_fifo_almost_full)? (th_comp == 6) && !_axi_a_read_req_fifo_almost_full && !_axi_a_read_req_fifo_almost_full : 0;
  localparam _tmp_26 = 1;
  wire [_tmp_26-1:0] _tmp_27;
  assign _tmp_27 = !_axi_a_read_req_fifo_almost_full;
  reg [_tmp_26-1:0] __tmp_27_1;
  reg [32-1:0] _axi_a_read_data_fsm;
  localparam _axi_a_read_data_fsm_init = 0;
  assign _axi_a_read_req_fifo_deq = ((_axi_a_read_data_fsm == 0) && (!_axi_a_read_data_busy && !_axi_a_read_req_fifo_empty && (_axi_a_read_op_sel_fifo == 1)) && !_axi_a_read_req_fifo_empty)? 1 : 0;
  reg [32-1:0] write_burst_fsm_0;
  localparam write_burst_fsm_0_init = 0;
  reg [10-1:0] write_burst_addr_28;
  reg [10-1:0] write_burst_stride_29;
  reg [33-1:0] write_burst_length_30;
  reg write_burst_done_31;
  assign ram_a_1_addr = ((write_burst_fsm_0 == 1) && axi_a_tvalid)? write_burst_addr_28 : 'hx;
  assign ram_a_1_wdata = ((write_burst_fsm_0 == 1) && axi_a_tvalid)? axi_a_tdata : 'hx;
  assign ram_a_1_wenable = ((write_burst_fsm_0 == 1) && axi_a_tvalid)? 1'd1 : 0;
  assign ram_a_1_enable = ((write_burst_fsm_0 == 1) && axi_a_tvalid)? 1'd1 : 0;
  assign axi_a_tready = _axi_a_read_data_fsm == 2;
  wire [8-1:0] pack_read_req_op_sel_32;
  wire [32-1:0] pack_read_req_local_addr_33;
  wire [32-1:0] pack_read_req_local_stride_34;
  wire [33-1:0] pack_read_req_local_size_35;
  assign pack_read_req_op_sel_32 = 1;
  assign pack_read_req_local_addr_33 = _th_comp_offset_1;
  assign pack_read_req_local_stride_34 = 1;
  assign pack_read_req_local_size_35 = _th_comp_size_0;
  wire [105-1:0] pack_read_req_packed_36;
  assign pack_read_req_packed_36 = { pack_read_req_op_sel_32, pack_read_req_local_addr_33, pack_read_req_local_stride_34, pack_read_req_local_size_35 };
  assign _axi_b_read_req_fifo_wdata = ((th_comp == 8) && !_axi_b_read_req_fifo_almost_full)? pack_read_req_packed_36 : 'hx;
  assign _axi_b_read_req_fifo_enq = ((th_comp == 8) && !_axi_b_read_req_fifo_almost_full)? (th_comp == 8) && !_axi_b_read_req_fifo_almost_full && !_axi_b_read_req_fifo_almost_full : 0;
  localparam _tmp_37 = 1;
  wire [_tmp_37-1:0] _tmp_38;
  assign _tmp_38 = !_axi_b_read_req_fifo_almost_full;
  reg [_tmp_37-1:0] __tmp_38_1;
  reg [32-1:0] _axi_b_read_data_fsm;
  localparam _axi_b_read_data_fsm_init = 0;
  assign _axi_b_read_req_fifo_deq = ((_axi_b_read_data_fsm == 0) && (!_axi_b_read_data_busy && !_axi_b_read_req_fifo_empty && (_axi_b_read_op_sel_fifo == 1)) && !_axi_b_read_req_fifo_empty)? 1 : 0;
  reg [32-1:0] write_burst_fsm_1;
  localparam write_burst_fsm_1_init = 0;
  reg [10-1:0] write_burst_addr_39;
  reg [10-1:0] write_burst_stride_40;
  reg [33-1:0] write_burst_length_41;
  reg write_burst_done_42;
  assign ram_b_1_addr = ((write_burst_fsm_1 == 1) && axi_b_tvalid)? write_burst_addr_39 : 'hx;
  assign ram_b_1_wdata = ((write_burst_fsm_1 == 1) && axi_b_tvalid)? axi_b_tdata : 'hx;
  assign ram_b_1_wenable = ((write_burst_fsm_1 == 1) && axi_b_tvalid)? 1'd1 : 0;
  assign ram_b_1_enable = ((write_burst_fsm_1 == 1) && axi_b_tvalid)? 1'd1 : 0;
  assign axi_b_tready = _axi_b_read_data_fsm == 2;
  reg signed [32-1:0] _th_comp_size_2;
  reg signed [32-1:0] _th_comp_offset_3;
  wire signed [32-1:0] mystream_a_data;
  wire signed [32-1:0] mystream_b_data;
  reg __mystream_stream_ivalid_1;
  reg signed [32-1:0] _plus_data_2;
  wire signed [32-1:0] mystream_c_data;
  assign mystream_c_data = _plus_data_2;
  wire _set_flag_43;
  assign _set_flag_43 = th_comp == 11;
  assign ram_a_0_addr = (_mystream_stream_oready && _mystream_a_source_ram_renable && (_mystream_a_source_sel == 1))? _mystream_a_source_ram_raddr : 'hx;
  assign ram_a_0_enable = (_mystream_stream_oready && _mystream_a_source_ram_renable && (_mystream_a_source_sel == 1))? 1'd1 : 0;
  localparam _tmp_44 = 1;
  wire [_tmp_44-1:0] _tmp_45;
  assign _tmp_45 = _mystream_stream_oready && _mystream_a_source_ram_renable && (_mystream_a_source_sel == 1);
  reg [_tmp_44-1:0] __tmp_45_1;
  assign _mystream_a_source_ram_rdata = (_mystream_a_source_sel == 1)? ram_a_0_rdata : 'hx;
  reg signed [32-1:0] __variable_wdata_0;
  assign mystream_a_data = __variable_wdata_0;
  reg [32-1:0] _mystream_a_source_fsm_0;
  localparam _mystream_a_source_fsm_0_init = 0;
  wire _set_flag_46;
  assign _set_flag_46 = th_comp == 12;
  assign ram_b_0_addr = (_mystream_stream_oready && _mystream_b_source_ram_renable && (_mystream_b_source_sel == 2))? _mystream_b_source_ram_raddr : 'hx;
  assign ram_b_0_enable = (_mystream_stream_oready && _mystream_b_source_ram_renable && (_mystream_b_source_sel == 2))? 1'd1 : 0;
  localparam _tmp_47 = 1;
  wire [_tmp_47-1:0] _tmp_48;
  assign _tmp_48 = _mystream_stream_oready && _mystream_b_source_ram_renable && (_mystream_b_source_sel == 2);
  reg [_tmp_47-1:0] __tmp_48_1;
  assign _mystream_b_source_ram_rdata = (_mystream_b_source_sel == 2)? ram_b_0_rdata : 'hx;
  reg signed [32-1:0] __variable_wdata_1;
  assign mystream_b_data = __variable_wdata_1;
  reg [32-1:0] _mystream_b_source_fsm_1;
  localparam _mystream_b_source_fsm_1_init = 0;
  wire _set_flag_49;
  assign _set_flag_49 = th_comp == 13;
  reg _tmp_50;
  reg _tmp_51;
  reg _tmp_52;
  reg signed [32-1:0] _tmp_53;
  reg signed [32-1:0] _tmp_54;
  reg signed [32-1:0] _tmp_55;
  reg signed [32-1:0] _tmp_56;
  reg signed [32-1:0] _tmp_57;
  reg signed [32-1:0] _tmp_58;
  assign ram_c_0_addr = (_mystream_stream_oready && _mystream_c_sink_wenable && (_mystream_c_sink_sel == 3))? _mystream_c_sink_waddr : 'hx;
  assign ram_c_0_wdata = (_mystream_stream_oready && _mystream_c_sink_wenable && (_mystream_c_sink_sel == 3))? _mystream_c_sink_wdata : 'hx;
  assign ram_c_0_wenable = (_mystream_stream_oready && _mystream_c_sink_wenable && (_mystream_c_sink_sel == 3))? 1'd1 : 0;
  assign ram_c_0_enable = (_mystream_stream_oready && _mystream_c_sink_wenable && (_mystream_c_sink_sel == 3))? 1'd1 : 0;
  reg [32-1:0] _mystream_c_sink_fsm_2;
  localparam _mystream_c_sink_fsm_2_init = 0;
  wire _set_flag_59;
  assign _set_flag_59 = th_comp == 14;
  assign _mystream_run_flag = (_set_flag_59)? 1 : 0;
  reg _tmp_60;
  reg _tmp_61;
  reg _tmp_62;
  assign _mystream_source_stop = _mystream_stream_oready && (_mystream_a_idle && _mystream_b_idle && (_mystream_fsm == 3));
  localparam _tmp_63 = 1;
  wire [_tmp_63-1:0] _tmp_64;
  assign _tmp_64 = _mystream_a_idle && _mystream_b_idle && (_mystream_fsm == 3);
  reg [_tmp_63-1:0] _tmp_65;
  reg _tmp_66;
  reg _tmp_67;
  reg _tmp_68;
  assign _mystream_sink_start = _tmp_68;
  reg _tmp_69;
  reg _tmp_70;
  reg _tmp_71;
  assign _mystream_sink_stop = _tmp_71;
  reg _tmp_72;
  reg _tmp_73;
  reg _tmp_74;
  assign _mystream_sink_busy = _tmp_74;
  reg _tmp_75;
  assign _mystream_busy = _mystream_source_busy || _mystream_sink_busy || _mystream_busy_reg;
  wire [8-1:0] pack_write_req_op_sel_76;
  wire [32-1:0] pack_write_req_local_addr_77;
  wire [32-1:0] pack_write_req_local_stride_78;
  wire [33-1:0] pack_write_req_local_size_79;
  assign pack_write_req_op_sel_76 = 1;
  assign pack_write_req_local_addr_77 = _th_comp_offset_1;
  assign pack_write_req_local_stride_78 = 1;
  assign pack_write_req_local_size_79 = _th_comp_size_0;
  wire [105-1:0] pack_write_req_packed_80;
  assign pack_write_req_packed_80 = { pack_write_req_op_sel_76, pack_write_req_local_addr_77, pack_write_req_local_stride_78, pack_write_req_local_size_79 };
  assign _axi_c_write_req_fifo_wdata = ((th_comp == 17) && !_axi_c_write_req_fifo_almost_full)? pack_write_req_packed_80 : 'hx;
  assign _axi_c_write_req_fifo_enq = ((th_comp == 17) && !_axi_c_write_req_fifo_almost_full)? (th_comp == 17) && !_axi_c_write_req_fifo_almost_full && !_axi_c_write_req_fifo_almost_full : 0;
  localparam _tmp_81 = 1;
  wire [_tmp_81-1:0] _tmp_82;
  assign _tmp_82 = !_axi_c_write_req_fifo_almost_full;
  reg [_tmp_81-1:0] __tmp_82_1;
  reg [32-1:0] _axi_c_write_data_fsm;
  localparam _axi_c_write_data_fsm_init = 0;
  assign _axi_c_write_req_fifo_deq = ((_axi_c_write_data_fsm == 0) && (!_axi_c_write_data_busy && !_axi_c_write_req_fifo_empty && (_axi_c_write_op_sel_fifo == 1)) && !_axi_c_write_req_fifo_empty)? 1 : 0;
  reg [32-1:0] read_burst_fsm_2;
  localparam read_burst_fsm_2_init = 0;
  reg [10-1:0] read_burst_addr_83;
  reg [10-1:0] read_burst_stride_84;
  reg [33-1:0] read_burst_length_85;
  reg read_burst_rvalid_86;
  reg read_burst_rlast_87;
  assign ram_c_1_addr = ((read_burst_fsm_2 == 1) && (!read_burst_rvalid_86 || (axi_c_tready || !axi_c_tvalid)))? read_burst_addr_83 : 'hx;
  assign ram_c_1_enable = ((read_burst_fsm_2 == 1) && (!read_burst_rvalid_86 || (axi_c_tready || !axi_c_tvalid)))? 1'd1 : 0;
  localparam _tmp_88 = 1;
  wire [_tmp_88-1:0] _tmp_89;
  assign _tmp_89 = (read_burst_fsm_2 == 1) && (!read_burst_rvalid_86 || (axi_c_tready || !axi_c_tvalid));
  reg [_tmp_88-1:0] __tmp_89_1;
  wire [32-1:0] read_burst_rdata_90;
  assign read_burst_rdata_90 = ram_c_1_rdata;
  reg _axi_c_cond_0_1;

  always @(posedge CLK) begin
    if(RST) begin
      _axi_a_read_data_busy <= 0;
      _axi_a_read_op_sel_buf <= 0;
      _axi_a_read_local_addr_buf <= 0;
      _axi_a_read_local_stride_buf <= 0;
      _axi_a_read_local_size_buf <= 0;
    end else begin
      if((_axi_a_read_data_fsm == 0) && (!_axi_a_read_data_busy && !_axi_a_read_req_fifo_empty && (_axi_a_read_op_sel_fifo == 1))) begin
        _axi_a_read_data_busy <= 1;
        _axi_a_read_op_sel_buf <= _axi_a_read_op_sel_fifo;
        _axi_a_read_local_addr_buf <= _axi_a_read_local_addr_fifo;
        _axi_a_read_local_stride_buf <= _axi_a_read_local_stride_fifo;
        _axi_a_read_local_size_buf <= _axi_a_read_local_size_fifo;
      end 
      if((_axi_a_read_data_fsm == 2) && axi_a_tvalid) begin
        _axi_a_read_local_size_buf <= _axi_a_read_local_size_buf - 1;
      end 
      if((_axi_a_read_data_fsm == 2) && axi_a_tvalid && (_axi_a_read_local_size_buf <= 1)) begin
        _axi_a_read_data_busy <= 0;
      end 
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      count__axi_a_read_req_fifo <= 0;
      __tmp_27_1 <= 0;
    end else begin
      if(_axi_a_read_req_fifo_enq && !_axi_a_read_req_fifo_full && (_axi_a_read_req_fifo_deq && !_axi_a_read_req_fifo_empty)) begin
        count__axi_a_read_req_fifo <= count__axi_a_read_req_fifo;
      end else if(_axi_a_read_req_fifo_enq && !_axi_a_read_req_fifo_full) begin
        count__axi_a_read_req_fifo <= count__axi_a_read_req_fifo + 1;
      end else if(_axi_a_read_req_fifo_deq && !_axi_a_read_req_fifo_empty) begin
        count__axi_a_read_req_fifo <= count__axi_a_read_req_fifo - 1;
      end 
      __tmp_27_1 <= _tmp_27;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      _axi_b_read_data_busy <= 0;
      _axi_b_read_op_sel_buf <= 0;
      _axi_b_read_local_addr_buf <= 0;
      _axi_b_read_local_stride_buf <= 0;
      _axi_b_read_local_size_buf <= 0;
    end else begin
      if((_axi_b_read_data_fsm == 0) && (!_axi_b_read_data_busy && !_axi_b_read_req_fifo_empty && (_axi_b_read_op_sel_fifo == 1))) begin
        _axi_b_read_data_busy <= 1;
        _axi_b_read_op_sel_buf <= _axi_b_read_op_sel_fifo;
        _axi_b_read_local_addr_buf <= _axi_b_read_local_addr_fifo;
        _axi_b_read_local_stride_buf <= _axi_b_read_local_stride_fifo;
        _axi_b_read_local_size_buf <= _axi_b_read_local_size_fifo;
      end 
      if((_axi_b_read_data_fsm == 2) && axi_b_tvalid) begin
        _axi_b_read_local_size_buf <= _axi_b_read_local_size_buf - 1;
      end 
      if((_axi_b_read_data_fsm == 2) && axi_b_tvalid && (_axi_b_read_local_size_buf <= 1)) begin
        _axi_b_read_data_busy <= 0;
      end 
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      count__axi_b_read_req_fifo <= 0;
      __tmp_38_1 <= 0;
    end else begin
      if(_axi_b_read_req_fifo_enq && !_axi_b_read_req_fifo_full && (_axi_b_read_req_fifo_deq && !_axi_b_read_req_fifo_empty)) begin
        count__axi_b_read_req_fifo <= count__axi_b_read_req_fifo;
      end else if(_axi_b_read_req_fifo_enq && !_axi_b_read_req_fifo_full) begin
        count__axi_b_read_req_fifo <= count__axi_b_read_req_fifo + 1;
      end else if(_axi_b_read_req_fifo_deq && !_axi_b_read_req_fifo_empty) begin
        count__axi_b_read_req_fifo <= count__axi_b_read_req_fifo - 1;
      end 
      __tmp_38_1 <= _tmp_38;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      axi_c_tdata <= 0;
      axi_c_tvalid <= 0;
      axi_c_tlast <= 0;
      _axi_c_cond_0_1 <= 0;
    end else begin
      if(_axi_c_cond_0_1) begin
        axi_c_tvalid <= 0;
        axi_c_tlast <= 0;
      end 
      if((_axi_c_write_op_sel_buf == 1) && read_burst_rvalid_86 && (axi_c_tready || !axi_c_tvalid) && (axi_c_tready || !axi_c_tvalid)) begin
        axi_c_tdata <= read_burst_rdata_90;
        axi_c_tvalid <= 1;
        axi_c_tlast <= read_burst_rlast_87;
      end 
      _axi_c_cond_0_1 <= 1;
      if(axi_c_tvalid && !axi_c_tready) begin
        axi_c_tvalid <= axi_c_tvalid;
        axi_c_tlast <= axi_c_tlast;
      end 
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      _axi_c_write_data_busy <= 0;
      _axi_c_write_op_sel_buf <= 0;
      _axi_c_write_local_addr_buf <= 0;
      _axi_c_write_local_stride_buf <= 0;
      _axi_c_write_size_buf <= 0;
    end else begin
      if((_axi_c_write_data_fsm == 0) && (!_axi_c_write_data_busy && !_axi_c_write_req_fifo_empty && (_axi_c_write_op_sel_fifo == 1))) begin
        _axi_c_write_data_busy <= 1;
        _axi_c_write_op_sel_buf <= _axi_c_write_op_sel_fifo;
        _axi_c_write_local_addr_buf <= _axi_c_write_local_addr_fifo;
        _axi_c_write_local_stride_buf <= _axi_c_write_local_stride_fifo;
        _axi_c_write_size_buf <= _axi_c_write_size_fifo;
      end 
      if((_axi_c_write_data_fsm == 2) && ((_axi_c_write_op_sel_buf == 1) && read_burst_rvalid_86 && (axi_c_tready || !axi_c_tvalid))) begin
        _axi_c_write_size_buf <= _axi_c_write_size_buf - 1;
      end 
      if((_axi_c_write_data_fsm == 2) && ((_axi_c_write_op_sel_buf == 1) && read_burst_rvalid_86 && (axi_c_tready || !axi_c_tvalid)) && read_burst_rlast_87) begin
        _axi_c_write_data_busy <= 0;
      end 
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      count__axi_c_write_req_fifo <= 0;
      __tmp_82_1 <= 0;
    end else begin
      if(_axi_c_write_req_fifo_enq && !_axi_c_write_req_fifo_full && (_axi_c_write_req_fifo_deq && !_axi_c_write_req_fifo_empty)) begin
        count__axi_c_write_req_fifo <= count__axi_c_write_req_fifo;
      end else if(_axi_c_write_req_fifo_enq && !_axi_c_write_req_fifo_full) begin
        count__axi_c_write_req_fifo <= count__axi_c_write_req_fifo + 1;
      end else if(_axi_c_write_req_fifo_deq && !_axi_c_write_req_fifo_empty) begin
        count__axi_c_write_req_fifo <= count__axi_c_write_req_fifo - 1;
      end 
      __tmp_82_1 <= _tmp_82;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      saxi_rdata <= 0;
      saxi_rvalid <= 0;
      _saxi_rdata_cond_0_1 <= 0;
    end else begin
      if(_saxi_rdata_cond_0_1) begin
        saxi_rvalid <= 0;
      end 
      if((_saxi_register_fsm == 1) && (saxi_rready || !saxi_rvalid)) begin
        saxi_rdata <= axislite_rdata_18;
        saxi_rvalid <= 1;
      end 
      _saxi_rdata_cond_0_1 <= 1;
      if(saxi_rvalid && !saxi_rready) begin
        saxi_rvalid <= saxi_rvalid;
      end 
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      saxi_bvalid <= 0;
      prev_awvalid_15 <= 0;
      prev_arvalid_16 <= 0;
      writevalid_13 <= 0;
      readvalid_14 <= 0;
      addr_12 <= 0;
      _saxi_register_0 <= 0;
      _saxi_flag_0 <= 0;
      _saxi_register_1 <= 0;
      _saxi_flag_1 <= 0;
      _saxi_register_2 <= 0;
      _saxi_flag_2 <= 0;
      _saxi_register_3 <= 0;
      _saxi_flag_3 <= 0;
    end else begin
      if(saxi_bvalid && saxi_bready) begin
        saxi_bvalid <= 0;
      end 
      if(saxi_wvalid && saxi_wready) begin
        saxi_bvalid <= 1;
      end 
      prev_awvalid_15 <= saxi_awvalid;
      prev_arvalid_16 <= saxi_arvalid;
      writevalid_13 <= 0;
      readvalid_14 <= 0;
      if(saxi_awready && saxi_awvalid && !saxi_bvalid) begin
        addr_12 <= saxi_awaddr;
        writevalid_13 <= 1;
      end else if(saxi_arready && saxi_arvalid) begin
        addr_12 <= saxi_araddr;
        readvalid_14 <= 1;
      end 
      if((_saxi_register_fsm == 1) && (saxi_rready || !saxi_rvalid) && axislite_flag_19 && (axis_maskaddr_17 == 0)) begin
        _saxi_register_0 <= axislite_resetval_20;
        _saxi_flag_0 <= 0;
      end 
      if((_saxi_register_fsm == 1) && (saxi_rready || !saxi_rvalid) && axislite_flag_19 && (axis_maskaddr_17 == 1)) begin
        _saxi_register_1 <= axislite_resetval_20;
        _saxi_flag_1 <= 0;
      end 
      if((_saxi_register_fsm == 1) && (saxi_rready || !saxi_rvalid) && axislite_flag_19 && (axis_maskaddr_17 == 2)) begin
        _saxi_register_2 <= axislite_resetval_20;
        _saxi_flag_2 <= 0;
      end 
      if((_saxi_register_fsm == 1) && (saxi_rready || !saxi_rvalid) && axislite_flag_19 && (axis_maskaddr_17 == 3)) begin
        _saxi_register_3 <= axislite_resetval_20;
        _saxi_flag_3 <= 0;
      end 
      if((_saxi_register_fsm == 3) && saxi_wvalid && (axis_maskaddr_17 == 0)) begin
        _saxi_register_0 <= saxi_wdata;
      end 
      if((_saxi_register_fsm == 3) && saxi_wvalid && (axis_maskaddr_17 == 1)) begin
        _saxi_register_1 <= saxi_wdata;
      end 
      if((_saxi_register_fsm == 3) && saxi_wvalid && (axis_maskaddr_17 == 2)) begin
        _saxi_register_2 <= saxi_wdata;
      end 
      if((_saxi_register_fsm == 3) && saxi_wvalid && (axis_maskaddr_17 == 3)) begin
        _saxi_register_3 <= saxi_wdata;
      end 
      if((_saxi_register_0 == 1) && (th_comp == 2) && 1) begin
        _saxi_register_0 <= 0;
      end 
      if((_saxi_register_0 == 1) && (th_comp == 2) && 0) begin
        _saxi_register_1 <= 0;
      end 
      if((_saxi_register_0 == 1) && (th_comp == 2) && 0) begin
        _saxi_register_2 <= 0;
      end 
      if((_saxi_register_0 == 1) && (th_comp == 2) && 0) begin
        _saxi_register_3 <= 0;
      end 
      if((th_comp == 3) && 0) begin
        _saxi_register_0 <= 1;
        _saxi_flag_0 <= 0;
      end 
      if((th_comp == 3) && 1) begin
        _saxi_register_1 <= 1;
        _saxi_flag_1 <= 0;
      end 
      if((th_comp == 3) && 0) begin
        _saxi_register_2 <= 1;
        _saxi_flag_2 <= 0;
      end 
      if((th_comp == 3) && 0) begin
        _saxi_register_3 <= 1;
        _saxi_flag_3 <= 0;
      end 
      if((th_comp == 19) && 0) begin
        _saxi_register_0 <= 0;
        _saxi_flag_0 <= 0;
      end 
      if((th_comp == 19) && 1) begin
        _saxi_register_1 <= 0;
        _saxi_flag_1 <= 0;
      end 
      if((th_comp == 19) && 0) begin
        _saxi_register_2 <= 0;
        _saxi_flag_2 <= 0;
      end 
      if((th_comp == 19) && 0) begin
        _saxi_register_3 <= 0;
        _saxi_flag_3 <= 0;
      end 
    end
  end

  localparam _saxi_register_fsm_1 = 1;
  localparam _saxi_register_fsm_2 = 2;
  localparam _saxi_register_fsm_3 = 3;
  localparam _saxi_register_fsm_4 = 4;

  always @(posedge CLK) begin
    if(RST) begin
      _saxi_register_fsm <= _saxi_register_fsm_init;
      axis_maskaddr_17 <= 0;
    end else begin
      case(_saxi_register_fsm)
        _saxi_register_fsm_init: begin
          if(readvalid_14 || writevalid_13) begin
            axis_maskaddr_17 <= (addr_12 >> _saxi_shift) & _saxi_mask;
          end 
          if(readvalid_14) begin
            _saxi_register_fsm <= _saxi_register_fsm_1;
          end 
          if(writevalid_13) begin
            _saxi_register_fsm <= _saxi_register_fsm_3;
          end 
        end
        _saxi_register_fsm_1: begin
          if(saxi_rready || !saxi_rvalid) begin
            _saxi_register_fsm <= _saxi_register_fsm_2;
          end 
        end
        _saxi_register_fsm_2: begin
          if(saxi_rready && saxi_rvalid) begin
            _saxi_register_fsm <= _saxi_register_fsm_init;
          end 
        end
        _saxi_register_fsm_3: begin
          if(saxi_wvalid) begin
            _saxi_register_fsm <= _saxi_register_fsm_4;
          end 
        end
        _saxi_register_fsm_4: begin
          if(saxi_bready && saxi_bvalid) begin
            _saxi_register_fsm <= _saxi_register_fsm_init;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      __tmp_45_1 <= 0;
    end else begin
      __tmp_45_1 <= _tmp_45;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      __tmp_48_1 <= 0;
    end else begin
      __tmp_48_1 <= _tmp_48;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      __tmp_89_1 <= 0;
    end else begin
      __tmp_89_1 <= _tmp_89;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      _mystream_a_source_ram_renable <= 0;
      _mystream_a_source_fifo_deq <= 0;
      _mystream_a_idle <= 1;
      _mystream_b_source_ram_renable <= 0;
      _mystream_b_source_fifo_deq <= 0;
      _mystream_b_idle <= 1;
      _mystream_c_sink_wenable <= 0;
      _mystream_c_sink_fifo_enq <= 0;
      __mystream_stream_ivalid_1 <= 0;
      _plus_data_2 <= 0;
      _mystream_a_source_mode <= 5'b0;
      _mystream_a_source_offset <= 0;
      _mystream_a_source_size <= 0;
      _mystream_a_source_stride <= 0;
      _mystream_a_source_sel <= 0;
      _mystream_a_source_offset_buf <= 0;
      _mystream_a_source_size_buf <= 0;
      _mystream_a_source_stride_buf <= 0;
      __variable_wdata_0 <= 0;
      _mystream_a_source_ram_raddr <= 0;
      _mystream_a_source_count <= 0;
      _mystream_b_source_mode <= 5'b0;
      _mystream_b_source_offset <= 0;
      _mystream_b_source_size <= 0;
      _mystream_b_source_stride <= 0;
      _mystream_b_source_sel <= 0;
      _mystream_b_source_offset_buf <= 0;
      _mystream_b_source_size_buf <= 0;
      _mystream_b_source_stride_buf <= 0;
      __variable_wdata_1 <= 0;
      _mystream_b_source_ram_raddr <= 0;
      _mystream_b_source_count <= 0;
      _tmp_50 <= 0;
      _tmp_51 <= 0;
      _tmp_52 <= 0;
      _tmp_53 <= 0;
      _tmp_54 <= 0;
      _tmp_55 <= 0;
      _tmp_56 <= 0;
      _tmp_57 <= 0;
      _tmp_58 <= 0;
      _mystream_c_sink_mode <= 5'b0;
      _mystream_c_sink_offset <= 0;
      _mystream_c_sink_size <= 0;
      _mystream_c_sink_stride <= 0;
      _mystream_c_sink_sel <= 0;
      _mystream_c_sink_offset_buf <= 0;
      _mystream_c_sink_size_buf <= 0;
      _mystream_c_sink_stride_buf <= 0;
      _mystream_c_sink_waddr <= 0;
      _mystream_c_sink_count <= 0;
      _mystream_c_sink_wdata <= 0;
      _tmp_60 <= 0;
      _tmp_61 <= 0;
      _tmp_62 <= 0;
      _tmp_65 <= 0;
      _tmp_66 <= 0;
      _tmp_67 <= 0;
      _tmp_68 <= 0;
      _tmp_69 <= 0;
      _tmp_70 <= 0;
      _tmp_71 <= 0;
      _tmp_72 <= 0;
      _tmp_73 <= 0;
      _tmp_74 <= 0;
      _tmp_75 <= 0;
      _mystream_busy_reg <= 0;
    end else begin
      if(_mystream_stream_oready) begin
        _mystream_a_source_ram_renable <= 0;
        _mystream_a_source_fifo_deq <= 0;
      end 
      _mystream_a_idle <= _mystream_a_idle;
      if(_mystream_stream_oready) begin
        _mystream_b_source_ram_renable <= 0;
        _mystream_b_source_fifo_deq <= 0;
      end 
      _mystream_b_idle <= _mystream_b_idle;
      if(_mystream_stream_oready) begin
        _mystream_c_sink_wenable <= 0;
        _mystream_c_sink_fifo_enq <= 0;
      end 
      if(_mystream_stream_oready) begin
        __mystream_stream_ivalid_1 <= _mystream_stream_ivalid;
      end 
      if(_mystream_stream_oready) begin
        _plus_data_2 <= mystream_a_data + mystream_b_data;
      end 
      if(_set_flag_43) begin
        _mystream_a_source_mode <= 5'b1;
        _mystream_a_source_offset <= _th_comp_offset_3;
        _mystream_a_source_size <= _th_comp_size_2;
        _mystream_a_source_stride <= 1;
      end 
      if(_set_flag_43) begin
        _mystream_a_source_sel <= 1;
      end 
      if(_mystream_source_start && _mystream_a_source_mode & 5'b1 && _mystream_stream_oready) begin
        _mystream_a_source_offset_buf <= _mystream_a_source_offset;
        _mystream_a_source_size_buf <= _mystream_a_source_size;
        _mystream_a_source_stride_buf <= _mystream_a_source_stride;
      end 
      if(_mystream_stream_oready && _mystream_source_busy && _mystream_is_root) begin
        __variable_wdata_0 <= _mystream_a_source_ram_rdata;
      end 
      if((_mystream_a_source_fsm_0 == 1) && _mystream_stream_oready) begin
        _mystream_a_idle <= 0;
        _mystream_a_source_ram_raddr <= _mystream_a_source_offset_buf;
        _mystream_a_source_ram_renable <= 1;
        _mystream_a_source_count <= _mystream_a_source_size_buf;
      end 
      if((_mystream_a_source_fsm_0 == 2) && _mystream_stream_oready) begin
        _mystream_a_source_ram_raddr <= _mystream_a_source_ram_raddr + _mystream_a_source_stride_buf;
        _mystream_a_source_ram_renable <= 1;
        _mystream_a_source_count <= _mystream_a_source_count - 1;
      end 
      if((_mystream_a_source_fsm_0 == 2) && (_mystream_a_source_count == 1) && _mystream_stream_oready) begin
        _mystream_a_source_ram_renable <= 0;
        _mystream_a_idle <= 1;
      end 
      if((_mystream_a_source_fsm_0 == 2) && _mystream_source_stop && _mystream_stream_oready) begin
        _mystream_a_source_ram_renable <= 0;
        _mystream_a_idle <= 1;
      end 
      if(_set_flag_46) begin
        _mystream_b_source_mode <= 5'b1;
        _mystream_b_source_offset <= _th_comp_offset_3;
        _mystream_b_source_size <= _th_comp_size_2;
        _mystream_b_source_stride <= 1;
      end 
      if(_set_flag_46) begin
        _mystream_b_source_sel <= 2;
      end 
      if(_mystream_source_start && _mystream_b_source_mode & 5'b1 && _mystream_stream_oready) begin
        _mystream_b_source_offset_buf <= _mystream_b_source_offset;
        _mystream_b_source_size_buf <= _mystream_b_source_size;
        _mystream_b_source_stride_buf <= _mystream_b_source_stride;
      end 
      if(_mystream_stream_oready && _mystream_source_busy && _mystream_is_root) begin
        __variable_wdata_1 <= _mystream_b_source_ram_rdata;
      end 
      if((_mystream_b_source_fsm_1 == 1) && _mystream_stream_oready) begin
        _mystream_b_idle <= 0;
        _mystream_b_source_ram_raddr <= _mystream_b_source_offset_buf;
        _mystream_b_source_ram_renable <= 1;
        _mystream_b_source_count <= _mystream_b_source_size_buf;
      end 
      if((_mystream_b_source_fsm_1 == 2) && _mystream_stream_oready) begin
        _mystream_b_source_ram_raddr <= _mystream_b_source_ram_raddr + _mystream_b_source_stride_buf;
        _mystream_b_source_ram_renable <= 1;
        _mystream_b_source_count <= _mystream_b_source_count - 1;
      end 
      if((_mystream_b_source_fsm_1 == 2) && (_mystream_b_source_count == 1) && _mystream_stream_oready) begin
        _mystream_b_source_ram_renable <= 0;
        _mystream_b_idle <= 1;
      end 
      if((_mystream_b_source_fsm_1 == 2) && _mystream_source_stop && _mystream_stream_oready) begin
        _mystream_b_source_ram_renable <= 0;
        _mystream_b_idle <= 1;
      end 
      if(_mystream_stream_oready) begin
        _tmp_50 <= _set_flag_49;
      end 
      if(_mystream_stream_oready) begin
        _tmp_51 <= _tmp_50;
      end 
      if(_mystream_stream_oready) begin
        _tmp_52 <= _tmp_51;
      end 
      if(_mystream_stream_oready) begin
        _tmp_53 <= _th_comp_offset_3;
      end 
      if(_mystream_stream_oready) begin
        _tmp_54 <= _tmp_53;
      end 
      if(_mystream_stream_oready) begin
        _tmp_55 <= _tmp_54;
      end 
      if(_mystream_stream_oready) begin
        _tmp_56 <= _th_comp_size_2;
      end 
      if(_mystream_stream_oready) begin
        _tmp_57 <= _tmp_56;
      end 
      if(_mystream_stream_oready) begin
        _tmp_58 <= _tmp_57;
      end 
      if(_tmp_52) begin
        _mystream_c_sink_mode <= 5'b1;
        _mystream_c_sink_offset <= _tmp_55;
        _mystream_c_sink_size <= _tmp_58;
        _mystream_c_sink_stride <= 1;
      end 
      if(_tmp_52) begin
        _mystream_c_sink_sel <= 3;
      end 
      if(_mystream_sink_start && _mystream_c_sink_mode & 5'b1 && _mystream_stream_oready) begin
        _mystream_c_sink_offset_buf <= _mystream_c_sink_offset;
        _mystream_c_sink_size_buf <= _mystream_c_sink_size;
        _mystream_c_sink_stride_buf <= _mystream_c_sink_stride;
      end 
      if((_mystream_c_sink_fsm_2 == 1) && _mystream_stream_oready) begin
        _mystream_c_sink_waddr <= _mystream_c_sink_offset_buf - _mystream_c_sink_stride_buf;
        _mystream_c_sink_count <= _mystream_c_sink_size_buf;
      end 
      if((_mystream_c_sink_fsm_2 == 2) && _mystream_stream_oready) begin
        _mystream_c_sink_waddr <= _mystream_c_sink_waddr + _mystream_c_sink_stride_buf;
        _mystream_c_sink_wdata <= mystream_c_data;
        _mystream_c_sink_wenable <= 1;
        _mystream_c_sink_count <= _mystream_c_sink_count - 1;
      end 
      if(_mystream_stream_oready) begin
        _tmp_60 <= _mystream_source_start;
      end 
      if(_mystream_stream_oready) begin
        _tmp_61 <= _tmp_60;
      end 
      if(_mystream_stream_oready) begin
        _tmp_62 <= _tmp_61;
      end 
      if(_mystream_stream_oready) begin
        _tmp_65 <= _tmp_64;
      end 
      if(_mystream_stream_oready) begin
        _tmp_66 <= _mystream_source_start;
      end 
      if(_mystream_stream_oready) begin
        _tmp_67 <= _tmp_66;
      end 
      if(_mystream_stream_oready) begin
        _tmp_68 <= _tmp_67;
      end 
      if(_mystream_stream_oready) begin
        _tmp_69 <= _mystream_source_stop;
      end 
      if(_mystream_stream_oready) begin
        _tmp_70 <= _tmp_69;
      end 
      if(_mystream_stream_oready) begin
        _tmp_71 <= _tmp_70;
      end 
      if(_mystream_stream_oready) begin
        _tmp_72 <= _mystream_source_busy;
      end 
      if(_mystream_stream_oready) begin
        _tmp_73 <= _tmp_72;
      end 
      if(_mystream_stream_oready) begin
        _tmp_74 <= _tmp_73;
      end 
      if(_mystream_stream_oready) begin
        _tmp_75 <= _mystream_sink_busy;
      end 
      if(!_mystream_sink_busy && _tmp_75) begin
        _mystream_busy_reg <= 0;
      end 
      if(_mystream_source_busy) begin
        _mystream_busy_reg <= 1;
      end 
    end
  end

  localparam _mystream_fsm_1 = 1;
  localparam _mystream_fsm_2 = 2;
  localparam _mystream_fsm_3 = 3;

  always @(posedge CLK) begin
    if(RST) begin
      _mystream_fsm <= _mystream_fsm_init;
      _mystream_source_start <= 0;
      _mystream_source_busy <= 0;
      _mystream_stream_ivalid <= 0;
    end else begin
      if(_mystream_stream_oready && _tmp_62) begin
        _mystream_stream_ivalid <= 1;
      end 
      if(_mystream_stream_oready && _tmp_65) begin
        _mystream_stream_ivalid <= 0;
      end 
      case(_mystream_fsm)
        _mystream_fsm_init: begin
          if(_mystream_run_flag) begin
            _mystream_source_start <= 1;
          end 
          if(_mystream_run_flag) begin
            _mystream_fsm <= _mystream_fsm_1;
          end 
        end
        _mystream_fsm_1: begin
          if(_mystream_source_start && _mystream_stream_oready) begin
            _mystream_source_start <= 0;
            _mystream_source_busy <= 1;
          end 
          if(_mystream_source_start && _mystream_stream_oready) begin
            _mystream_fsm <= _mystream_fsm_2;
          end 
        end
        _mystream_fsm_2: begin
          if(_mystream_stream_oready) begin
            _mystream_fsm <= _mystream_fsm_3;
          end 
        end
        _mystream_fsm_3: begin
          if(_mystream_stream_oready && (_mystream_a_idle && _mystream_b_idle && (_mystream_fsm == 3))) begin
            _mystream_source_busy <= 0;
          end 
          if(_mystream_stream_oready && (_mystream_a_idle && _mystream_b_idle && (_mystream_fsm == 3)) && _mystream_run_flag) begin
            _mystream_source_start <= 1;
          end 
          if(_mystream_stream_oready && (_mystream_a_idle && _mystream_b_idle && (_mystream_fsm == 3))) begin
            _mystream_fsm <= _mystream_fsm_init;
          end 
          if(_mystream_stream_oready && (_mystream_a_idle && _mystream_b_idle && (_mystream_fsm == 3)) && _mystream_run_flag) begin
            _mystream_fsm <= _mystream_fsm_1;
          end 
        end
      endcase
    end
  end

  localparam th_comp_1 = 1;
  localparam th_comp_2 = 2;
  localparam th_comp_3 = 3;
  localparam th_comp_4 = 4;
  localparam th_comp_5 = 5;
  localparam th_comp_6 = 6;
  localparam th_comp_7 = 7;
  localparam th_comp_8 = 8;
  localparam th_comp_9 = 9;
  localparam th_comp_10 = 10;
  localparam th_comp_11 = 11;
  localparam th_comp_12 = 12;
  localparam th_comp_13 = 13;
  localparam th_comp_14 = 14;
  localparam th_comp_15 = 15;
  localparam th_comp_16 = 16;
  localparam th_comp_17 = 17;
  localparam th_comp_18 = 18;
  localparam th_comp_19 = 19;
  localparam th_comp_20 = 20;
  localparam th_comp_21 = 21;
  localparam th_comp_22 = 22;

  always @(posedge CLK) begin
    if(RST) begin
      th_comp <= th_comp_init;
      _th_comp_size_0 <= 0;
      _th_comp_offset_1 <= 0;
      _th_comp_size_2 <= 0;
      _th_comp_offset_3 <= 0;
    end else begin
      case(th_comp)
        th_comp_init: begin
          th_comp <= th_comp_1;
        end
        th_comp_1: begin
          if(1) begin
            th_comp <= th_comp_2;
          end else begin
            th_comp <= th_comp_21;
          end
        end
        th_comp_2: begin
          if(_saxi_register_0 == 1) begin
            th_comp <= th_comp_3;
          end 
        end
        th_comp_3: begin
          th_comp <= th_comp_4;
        end
        th_comp_4: begin
          _th_comp_size_0 <= _saxi_register_2;
          th_comp <= th_comp_5;
        end
        th_comp_5: begin
          _th_comp_offset_1 <= 0;
          th_comp <= th_comp_6;
        end
        th_comp_6: begin
          if(!_axi_a_read_req_fifo_almost_full) begin
            th_comp <= th_comp_7;
          end 
        end
        th_comp_7: begin
          if(_axi_a_read_idle) begin
            th_comp <= th_comp_8;
          end 
        end
        th_comp_8: begin
          if(!_axi_b_read_req_fifo_almost_full) begin
            th_comp <= th_comp_9;
          end 
        end
        th_comp_9: begin
          if(_axi_b_read_idle) begin
            th_comp <= th_comp_10;
          end 
        end
        th_comp_10: begin
          _th_comp_size_2 <= _th_comp_size_0;
          _th_comp_offset_3 <= _th_comp_offset_1;
          th_comp <= th_comp_11;
        end
        th_comp_11: begin
          th_comp <= th_comp_12;
        end
        th_comp_12: begin
          th_comp <= th_comp_13;
        end
        th_comp_13: begin
          if(_mystream_stream_oready) begin
            th_comp <= th_comp_14;
          end 
        end
        th_comp_14: begin
          th_comp <= th_comp_15;
        end
        th_comp_15: begin
          if(_mystream_busy) begin
            th_comp <= th_comp_16;
          end 
        end
        th_comp_16: begin
          if(!_mystream_busy) begin
            th_comp <= th_comp_17;
          end 
        end
        th_comp_17: begin
          if(!_axi_c_write_req_fifo_almost_full) begin
            th_comp <= th_comp_18;
          end 
        end
        th_comp_18: begin
          if(_axi_c_write_idle) begin
            th_comp <= th_comp_19;
          end 
        end
        th_comp_19: begin
          th_comp <= th_comp_20;
        end
        th_comp_20: begin
          th_comp <= th_comp_1;
        end
        th_comp_21: begin
          $finish;
          th_comp <= th_comp_22;
        end
      endcase
    end
  end

  localparam _axi_a_read_data_fsm_1 = 1;
  localparam _axi_a_read_data_fsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      _axi_a_read_data_fsm <= _axi_a_read_data_fsm_init;
    end else begin
      case(_axi_a_read_data_fsm)
        _axi_a_read_data_fsm_init: begin
          if(!_axi_a_read_data_busy && !_axi_a_read_req_fifo_empty && (_axi_a_read_op_sel_fifo == 1)) begin
            _axi_a_read_data_fsm <= _axi_a_read_data_fsm_1;
          end 
        end
        _axi_a_read_data_fsm_1: begin
          _axi_a_read_data_fsm <= _axi_a_read_data_fsm_2;
        end
        _axi_a_read_data_fsm_2: begin
          if(axi_a_tvalid && (_axi_a_read_local_size_buf <= 1)) begin
            _axi_a_read_data_fsm <= _axi_a_read_data_fsm_init;
          end 
        end
      endcase
    end
  end

  localparam write_burst_fsm_0_1 = 1;

  always @(posedge CLK) begin
    if(RST) begin
      write_burst_fsm_0 <= write_burst_fsm_0_init;
      write_burst_addr_28 <= 0;
      write_burst_stride_29 <= 0;
      write_burst_length_30 <= 0;
      write_burst_done_31 <= 0;
    end else begin
      case(write_burst_fsm_0)
        write_burst_fsm_0_init: begin
          write_burst_addr_28 <= _axi_a_read_local_addr_buf;
          write_burst_stride_29 <= _axi_a_read_local_stride_buf;
          write_burst_length_30 <= _axi_a_read_local_size_buf;
          write_burst_done_31 <= 0;
          if((_axi_a_read_data_fsm == 1) && (_axi_a_read_op_sel_buf == 1) && (_axi_a_read_local_size_buf > 0)) begin
            write_burst_fsm_0 <= write_burst_fsm_0_1;
          end 
        end
        write_burst_fsm_0_1: begin
          if(axi_a_tvalid) begin
            write_burst_addr_28 <= write_burst_addr_28 + write_burst_stride_29;
            write_burst_length_30 <= write_burst_length_30 - 1;
            write_burst_done_31 <= 0;
          end 
          if(axi_a_tvalid && (write_burst_length_30 <= 1)) begin
            write_burst_done_31 <= 1;
          end 
          if(axi_a_tvalid && 0) begin
            write_burst_done_31 <= 1;
          end 
          if(axi_a_tvalid && (write_burst_length_30 <= 1)) begin
            write_burst_fsm_0 <= write_burst_fsm_0_init;
          end 
          if(axi_a_tvalid && 0) begin
            write_burst_fsm_0 <= write_burst_fsm_0_init;
          end 
          if(0) begin
            write_burst_fsm_0 <= write_burst_fsm_0_init;
          end 
        end
      endcase
    end
  end

  localparam _axi_b_read_data_fsm_1 = 1;
  localparam _axi_b_read_data_fsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      _axi_b_read_data_fsm <= _axi_b_read_data_fsm_init;
    end else begin
      case(_axi_b_read_data_fsm)
        _axi_b_read_data_fsm_init: begin
          if(!_axi_b_read_data_busy && !_axi_b_read_req_fifo_empty && (_axi_b_read_op_sel_fifo == 1)) begin
            _axi_b_read_data_fsm <= _axi_b_read_data_fsm_1;
          end 
        end
        _axi_b_read_data_fsm_1: begin
          _axi_b_read_data_fsm <= _axi_b_read_data_fsm_2;
        end
        _axi_b_read_data_fsm_2: begin
          if(axi_b_tvalid && (_axi_b_read_local_size_buf <= 1)) begin
            _axi_b_read_data_fsm <= _axi_b_read_data_fsm_init;
          end 
        end
      endcase
    end
  end

  localparam write_burst_fsm_1_1 = 1;

  always @(posedge CLK) begin
    if(RST) begin
      write_burst_fsm_1 <= write_burst_fsm_1_init;
      write_burst_addr_39 <= 0;
      write_burst_stride_40 <= 0;
      write_burst_length_41 <= 0;
      write_burst_done_42 <= 0;
    end else begin
      case(write_burst_fsm_1)
        write_burst_fsm_1_init: begin
          write_burst_addr_39 <= _axi_b_read_local_addr_buf;
          write_burst_stride_40 <= _axi_b_read_local_stride_buf;
          write_burst_length_41 <= _axi_b_read_local_size_buf;
          write_burst_done_42 <= 0;
          if((_axi_b_read_data_fsm == 1) && (_axi_b_read_op_sel_buf == 1) && (_axi_b_read_local_size_buf > 0)) begin
            write_burst_fsm_1 <= write_burst_fsm_1_1;
          end 
        end
        write_burst_fsm_1_1: begin
          if(axi_b_tvalid) begin
            write_burst_addr_39 <= write_burst_addr_39 + write_burst_stride_40;
            write_burst_length_41 <= write_burst_length_41 - 1;
            write_burst_done_42 <= 0;
          end 
          if(axi_b_tvalid && (write_burst_length_41 <= 1)) begin
            write_burst_done_42 <= 1;
          end 
          if(axi_b_tvalid && 0) begin
            write_burst_done_42 <= 1;
          end 
          if(axi_b_tvalid && (write_burst_length_41 <= 1)) begin
            write_burst_fsm_1 <= write_burst_fsm_1_init;
          end 
          if(axi_b_tvalid && 0) begin
            write_burst_fsm_1 <= write_burst_fsm_1_init;
          end 
          if(0) begin
            write_burst_fsm_1 <= write_burst_fsm_1_init;
          end 
        end
      endcase
    end
  end

  localparam _mystream_a_source_fsm_0_1 = 1;
  localparam _mystream_a_source_fsm_0_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      _mystream_a_source_fsm_0 <= _mystream_a_source_fsm_0_init;
    end else begin
      case(_mystream_a_source_fsm_0)
        _mystream_a_source_fsm_0_init: begin
          if(_mystream_source_start && _mystream_a_source_mode & 5'b1 && _mystream_stream_oready) begin
            _mystream_a_source_fsm_0 <= _mystream_a_source_fsm_0_1;
          end 
        end
        _mystream_a_source_fsm_0_1: begin
          if(_mystream_stream_oready) begin
            _mystream_a_source_fsm_0 <= _mystream_a_source_fsm_0_2;
          end 
        end
        _mystream_a_source_fsm_0_2: begin
          if((_mystream_a_source_count == 1) && _mystream_stream_oready) begin
            _mystream_a_source_fsm_0 <= _mystream_a_source_fsm_0_init;
          end 
          if(_mystream_source_stop && _mystream_stream_oready) begin
            _mystream_a_source_fsm_0 <= _mystream_a_source_fsm_0_init;
          end 
        end
      endcase
    end
  end

  localparam _mystream_b_source_fsm_1_1 = 1;
  localparam _mystream_b_source_fsm_1_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      _mystream_b_source_fsm_1 <= _mystream_b_source_fsm_1_init;
    end else begin
      case(_mystream_b_source_fsm_1)
        _mystream_b_source_fsm_1_init: begin
          if(_mystream_source_start && _mystream_b_source_mode & 5'b1 && _mystream_stream_oready) begin
            _mystream_b_source_fsm_1 <= _mystream_b_source_fsm_1_1;
          end 
        end
        _mystream_b_source_fsm_1_1: begin
          if(_mystream_stream_oready) begin
            _mystream_b_source_fsm_1 <= _mystream_b_source_fsm_1_2;
          end 
        end
        _mystream_b_source_fsm_1_2: begin
          if((_mystream_b_source_count == 1) && _mystream_stream_oready) begin
            _mystream_b_source_fsm_1 <= _mystream_b_source_fsm_1_init;
          end 
          if(_mystream_source_stop && _mystream_stream_oready) begin
            _mystream_b_source_fsm_1 <= _mystream_b_source_fsm_1_init;
          end 
        end
      endcase
    end
  end

  localparam _mystream_c_sink_fsm_2_1 = 1;
  localparam _mystream_c_sink_fsm_2_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      _mystream_c_sink_fsm_2 <= _mystream_c_sink_fsm_2_init;
    end else begin
      case(_mystream_c_sink_fsm_2)
        _mystream_c_sink_fsm_2_init: begin
          if(_mystream_sink_start && _mystream_c_sink_mode & 5'b1 && _mystream_stream_oready) begin
            _mystream_c_sink_fsm_2 <= _mystream_c_sink_fsm_2_1;
          end 
        end
        _mystream_c_sink_fsm_2_1: begin
          if(_mystream_stream_oready) begin
            _mystream_c_sink_fsm_2 <= _mystream_c_sink_fsm_2_2;
          end 
        end
        _mystream_c_sink_fsm_2_2: begin
          if((_mystream_c_sink_count == 1) && _mystream_stream_oready) begin
            _mystream_c_sink_fsm_2 <= _mystream_c_sink_fsm_2_init;
          end 
          if(_mystream_sink_stop && _mystream_stream_oready) begin
            _mystream_c_sink_fsm_2 <= _mystream_c_sink_fsm_2_init;
          end 
        end
      endcase
    end
  end

  localparam _axi_c_write_data_fsm_1 = 1;
  localparam _axi_c_write_data_fsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      _axi_c_write_data_fsm <= _axi_c_write_data_fsm_init;
    end else begin
      case(_axi_c_write_data_fsm)
        _axi_c_write_data_fsm_init: begin
          if(!_axi_c_write_data_busy && !_axi_c_write_req_fifo_empty && (_axi_c_write_op_sel_fifo == 1)) begin
            _axi_c_write_data_fsm <= _axi_c_write_data_fsm_1;
          end 
        end
        _axi_c_write_data_fsm_1: begin
          _axi_c_write_data_fsm <= _axi_c_write_data_fsm_2;
        end
        _axi_c_write_data_fsm_2: begin
          if((_axi_c_write_op_sel_buf == 1) && read_burst_rvalid_86 && (axi_c_tready || !axi_c_tvalid) && read_burst_rlast_87) begin
            _axi_c_write_data_fsm <= _axi_c_write_data_fsm_init;
          end 
        end
      endcase
    end
  end

  localparam read_burst_fsm_2_1 = 1;

  always @(posedge CLK) begin
    if(RST) begin
      read_burst_fsm_2 <= read_burst_fsm_2_init;
      read_burst_addr_83 <= 0;
      read_burst_stride_84 <= 0;
      read_burst_length_85 <= 0;
      read_burst_rvalid_86 <= 0;
      read_burst_rlast_87 <= 0;
    end else begin
      case(read_burst_fsm_2)
        read_burst_fsm_2_init: begin
          read_burst_addr_83 <= _axi_c_write_local_addr_buf;
          read_burst_stride_84 <= _axi_c_write_local_stride_buf;
          read_burst_length_85 <= _axi_c_write_size_buf;
          read_burst_rvalid_86 <= 0;
          read_burst_rlast_87 <= 0;
          if((_axi_c_write_data_fsm == 1) && (_axi_c_write_op_sel_buf == 1) && (_axi_c_write_size_buf > 0)) begin
            read_burst_fsm_2 <= read_burst_fsm_2_1;
          end 
        end
        read_burst_fsm_2_1: begin
          if((axi_c_tready || !axi_c_tvalid) && (read_burst_length_85 > 0)) begin
            read_burst_addr_83 <= read_burst_addr_83 + read_burst_stride_84;
            read_burst_length_85 <= read_burst_length_85 - 1;
            read_burst_rvalid_86 <= 1;
          end 
          if((axi_c_tready || !axi_c_tvalid) && (read_burst_length_85 <= 1)) begin
            read_burst_rlast_87 <= 1;
          end 
          if(read_burst_rlast_87 && read_burst_rvalid_86 && (axi_c_tready || !axi_c_tvalid)) begin
            read_burst_rvalid_86 <= 0;
            read_burst_rlast_87 <= 0;
          end 
          if(0) begin
            read_burst_rvalid_86 <= 0;
            read_burst_rlast_87 <= 0;
          end 
          if(read_burst_rlast_87 && read_burst_rvalid_86 && (axi_c_tready || !axi_c_tvalid)) begin
            read_burst_fsm_2 <= read_burst_fsm_2_init;
          end 
          if(0) begin
            read_burst_fsm_2 <= read_burst_fsm_2_init;
          end 
        end
      endcase
    end
  end


endmodule



module _axi_a_read_req_fifo
(
  input CLK,
  input RST,
  input _axi_a_read_req_fifo_enq,
  input [105-1:0] _axi_a_read_req_fifo_wdata,
  output _axi_a_read_req_fifo_full,
  output _axi_a_read_req_fifo_almost_full,
  input _axi_a_read_req_fifo_deq,
  output [105-1:0] _axi_a_read_req_fifo_rdata,
  output _axi_a_read_req_fifo_empty,
  output _axi_a_read_req_fifo_almost_empty
);

  reg [105-1:0] mem [0:8-1];
  reg [3-1:0] head;
  reg [3-1:0] tail;
  wire is_empty;
  wire is_almost_empty;
  wire is_full;
  wire is_almost_full;
  assign is_empty = head == tail;
  assign is_almost_empty = head == (tail + 1 & 7);
  assign is_full = (head + 1 & 7) == tail;
  assign is_almost_full = (head + 2 & 7) == tail;
  wire [105-1:0] rdata;
  assign _axi_a_read_req_fifo_full = is_full;
  assign _axi_a_read_req_fifo_almost_full = is_almost_full || is_full;
  assign _axi_a_read_req_fifo_empty = is_empty;
  assign _axi_a_read_req_fifo_almost_empty = is_almost_empty || is_empty;
  assign rdata = mem[tail];
  assign _axi_a_read_req_fifo_rdata = rdata;

  always @(posedge CLK) begin
    if(RST) begin
      head <= 0;
      tail <= 0;
    end else begin
      if(_axi_a_read_req_fifo_enq && !is_full) begin
        mem[head] <= _axi_a_read_req_fifo_wdata;
        head <= head + 1;
      end 
      if(_axi_a_read_req_fifo_deq && !is_empty) begin
        tail <= tail + 1;
      end 
    end
  end


endmodule



module _axi_b_read_req_fifo
(
  input CLK,
  input RST,
  input _axi_b_read_req_fifo_enq,
  input [105-1:0] _axi_b_read_req_fifo_wdata,
  output _axi_b_read_req_fifo_full,
  output _axi_b_read_req_fifo_almost_full,
  input _axi_b_read_req_fifo_deq,
  output [105-1:0] _axi_b_read_req_fifo_rdata,
  output _axi_b_read_req_fifo_empty,
  output _axi_b_read_req_fifo_almost_empty
);

  reg [105-1:0] mem [0:8-1];
  reg [3-1:0] head;
  reg [3-1:0] tail;
  wire is_empty;
  wire is_almost_empty;
  wire is_full;
  wire is_almost_full;
  assign is_empty = head == tail;
  assign is_almost_empty = head == (tail + 1 & 7);
  assign is_full = (head + 1 & 7) == tail;
  assign is_almost_full = (head + 2 & 7) == tail;
  wire [105-1:0] rdata;
  assign _axi_b_read_req_fifo_full = is_full;
  assign _axi_b_read_req_fifo_almost_full = is_almost_full || is_full;
  assign _axi_b_read_req_fifo_empty = is_empty;
  assign _axi_b_read_req_fifo_almost_empty = is_almost_empty || is_empty;
  assign rdata = mem[tail];
  assign _axi_b_read_req_fifo_rdata = rdata;

  always @(posedge CLK) begin
    if(RST) begin
      head <= 0;
      tail <= 0;
    end else begin
      if(_axi_b_read_req_fifo_enq && !is_full) begin
        mem[head] <= _axi_b_read_req_fifo_wdata;
        head <= head + 1;
      end 
      if(_axi_b_read_req_fifo_deq && !is_empty) begin
        tail <= tail + 1;
      end 
    end
  end


endmodule



module _axi_c_write_req_fifo
(
  input CLK,
  input RST,
  input _axi_c_write_req_fifo_enq,
  input [105-1:0] _axi_c_write_req_fifo_wdata,
  output _axi_c_write_req_fifo_full,
  output _axi_c_write_req_fifo_almost_full,
  input _axi_c_write_req_fifo_deq,
  output [105-1:0] _axi_c_write_req_fifo_rdata,
  output _axi_c_write_req_fifo_empty,
  output _axi_c_write_req_fifo_almost_empty
);

  reg [105-1:0] mem [0:8-1];
  reg [3-1:0] head;
  reg [3-1:0] tail;
  wire is_empty;
  wire is_almost_empty;
  wire is_full;
  wire is_almost_full;
  assign is_empty = head == tail;
  assign is_almost_empty = head == (tail + 1 & 7);
  assign is_full = (head + 1 & 7) == tail;
  assign is_almost_full = (head + 2 & 7) == tail;
  wire [105-1:0] rdata;
  assign _axi_c_write_req_fifo_full = is_full;
  assign _axi_c_write_req_fifo_almost_full = is_almost_full || is_full;
  assign _axi_c_write_req_fifo_empty = is_empty;
  assign _axi_c_write_req_fifo_almost_empty = is_almost_empty || is_empty;
  assign rdata = mem[tail];
  assign _axi_c_write_req_fifo_rdata = rdata;

  always @(posedge CLK) begin
    if(RST) begin
      head <= 0;
      tail <= 0;
    end else begin
      if(_axi_c_write_req_fifo_enq && !is_full) begin
        mem[head] <= _axi_c_write_req_fifo_wdata;
        head <= head + 1;
      end 
      if(_axi_c_write_req_fifo_deq && !is_empty) begin
        tail <= tail + 1;
      end 
    end
  end


endmodule



module ram_a
(
  input CLK,
  input [10-1:0] ram_a_0_addr,
  output [32-1:0] ram_a_0_rdata,
  input [32-1:0] ram_a_0_wdata,
  input ram_a_0_wenable,
  input ram_a_0_enable,
  input [10-1:0] ram_a_1_addr,
  output [32-1:0] ram_a_1_rdata,
  input [32-1:0] ram_a_1_wdata,
  input ram_a_1_wenable,
  input ram_a_1_enable
);

  reg [32-1:0] ram_a_0_rdata_out;
  assign ram_a_0_rdata = ram_a_0_rdata_out;
  reg [32-1:0] ram_a_1_rdata_out;
  assign ram_a_1_rdata = ram_a_1_rdata_out;
  reg [32-1:0] mem [0:1024-1];

  always @(posedge CLK) begin
    if(ram_a_0_enable) begin
      if(ram_a_0_wenable) begin
        mem[ram_a_0_addr] <= ram_a_0_wdata;
        ram_a_0_rdata_out <= ram_a_0_wdata;
      end else begin
        ram_a_0_rdata_out <= mem[ram_a_0_addr];
      end
    end 
  end


  always @(posedge CLK) begin
    if(ram_a_1_enable) begin
      if(ram_a_1_wenable) begin
        mem[ram_a_1_addr] <= ram_a_1_wdata;
        ram_a_1_rdata_out <= ram_a_1_wdata;
      end else begin
        ram_a_1_rdata_out <= mem[ram_a_1_addr];
      end
    end 
  end


endmodule



module ram_b
(
  input CLK,
  input [10-1:0] ram_b_0_addr,
  output [32-1:0] ram_b_0_rdata,
  input [32-1:0] ram_b_0_wdata,
  input ram_b_0_wenable,
  input ram_b_0_enable,
  input [10-1:0] ram_b_1_addr,
  output [32-1:0] ram_b_1_rdata,
  input [32-1:0] ram_b_1_wdata,
  input ram_b_1_wenable,
  input ram_b_1_enable
);

  reg [32-1:0] ram_b_0_rdata_out;
  assign ram_b_0_rdata = ram_b_0_rdata_out;
  reg [32-1:0] ram_b_1_rdata_out;
  assign ram_b_1_rdata = ram_b_1_rdata_out;
  reg [32-1:0] mem [0:1024-1];

  always @(posedge CLK) begin
    if(ram_b_0_enable) begin
      if(ram_b_0_wenable) begin
        mem[ram_b_0_addr] <= ram_b_0_wdata;
        ram_b_0_rdata_out <= ram_b_0_wdata;
      end else begin
        ram_b_0_rdata_out <= mem[ram_b_0_addr];
      end
    end 
  end


  always @(posedge CLK) begin
    if(ram_b_1_enable) begin
      if(ram_b_1_wenable) begin
        mem[ram_b_1_addr] <= ram_b_1_wdata;
        ram_b_1_rdata_out <= ram_b_1_wdata;
      end else begin
        ram_b_1_rdata_out <= mem[ram_b_1_addr];
      end
    end 
  end


endmodule



module ram_c
(
  input CLK,
  input [10-1:0] ram_c_0_addr,
  output [32-1:0] ram_c_0_rdata,
  input [32-1:0] ram_c_0_wdata,
  input ram_c_0_wenable,
  input ram_c_0_enable,
  input [10-1:0] ram_c_1_addr,
  output [32-1:0] ram_c_1_rdata,
  input [32-1:0] ram_c_1_wdata,
  input ram_c_1_wenable,
  input ram_c_1_enable
);

  reg [32-1:0] ram_c_0_rdata_out;
  assign ram_c_0_rdata = ram_c_0_rdata_out;
  reg [32-1:0] ram_c_1_rdata_out;
  assign ram_c_1_rdata = ram_c_1_rdata_out;
  reg [32-1:0] mem [0:1024-1];

  always @(posedge CLK) begin
    if(ram_c_0_enable) begin
      if(ram_c_0_wenable) begin
        mem[ram_c_0_addr] <= ram_c_0_wdata;
        ram_c_0_rdata_out <= ram_c_0_wdata;
      end else begin
        ram_c_0_rdata_out <= mem[ram_c_0_addr];
      end
    end 
  end


  always @(posedge CLK) begin
    if(ram_c_1_enable) begin
      if(ram_c_1_wenable) begin
        mem[ram_c_1_addr] <= ram_c_1_wdata;
        ram_c_1_rdata_out <= ram_c_1_wdata;
      end else begin
        ram_c_1_rdata_out <= mem[ram_c_1_addr];
      end
    end 
  end


endmodule
"""


def test(request):
    veriloggen.reset()

    simtype = request.config.getoption('--sim')

    code = thread_stream_axi_stream.run(filename=None, simtype=simtype,
                                        outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator

    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *
from veriloggen.thread import RAM


def mkMain(n=128, datawidth=32, numports=2):
    m = Module('main')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    addrwidth = int(math.log(n, 2)) * 2

    myram = RAM(m, 'myram', clk, rst, datawidth, addrwidth, 1)

    # example how to access RAM
    count = m.Reg('count', 32, initval=0)
    sum = m.Reg('sum', 32, initval=0)
    addr = m.Reg('addr', 32, initval=0)

    fsm = FSM(m, 'fsm', clk, rst)

    fsm(
        addr(0),
        count(0),
        sum(0)
    )

    fsm.goto_next()

    step = 16

    myram.write_rtl(addr, count, port=0, cond=fsm)

    fsm(
        addr.inc(),
        count.inc()
    )

    fsm.If(count == step - 1)(
        addr(0),
        count(0)
    )

    fsm.Then().goto_next()

    read_data, read_valid = myram.read_rtl(addr, port=0, cond=fsm)

    fsm(
        addr.inc(),
        count.inc()
    )

    fsm.If(read_valid)(
        sum(sum + read_data)
    )

    fsm.Then().Delay(1)(
        Systask('display', "sum=%d", sum)
    )

    fsm.If(count == step - 1)(
        addr(0),
        count(0)
    ).Then().goto_next()

    fsm.If(read_valid)(
        sum(sum + read_data)
    )

    fsm.Then().Delay(1)(
        Systask('display', "sum=%d", sum)
    )

    fsm.goto_next()

    fsm(
        Systask('display', "expected_sum=%d", (step - 1) * step // 2)
    )

    fsm.goto_next()

    fsm.make_always()

    return m


def mkTest():
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    vcd_name = os.path.splitext(os.path.basename(__file__))[0] + '.vcd'
    simulation.setup_waveform(m, uut, m.get_vars(), dumpfile=vcd_name)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m


if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    # sim.view_waveform()



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *


def mkLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', width, initval=0)
    count = m.Reg('count', 32, initval=0)

    """
    led = 0
    while True:
        count = 0
        while count < 1024:
            count += 1
        led += 1
    """

    fsm = FSM(m, 'fsm', clk, rst)

    fsm(led(0))
    fsm.goto_next()

    init = fsm.current
    fsm(count(0))
    fsm.goto_next()

    head = fsm.current
    cond = count < 1024
    fsm.goto_next(cond=cond)

    fsm(count.inc())
    fsm.goto_next()

    fsm(count.inc())
    fsm.goto(head)
    fsm.inc()

    tail = fsm.current
    fsm.goto_from(head, tail, cond=Not(cond))

    fsm(led.inc())
    fsm.goto(init)

    # fsm.make_always()
    # make_alway() is called when to_veirlog() is called.
    m.add_hook(fsm.make_always, args=(), kwargs={})

    return m


def mkTop():
    m = Module('top')
    led = mkLed()
    params = m.copy_params(led)
    ports = m.copy_ports(led)
    m.Instance(led, 'inst_' + led.name,
               connect_same_name(*params.values()),
               connect_same_name(*ports.values()))
    return m


if __name__ == '__main__':
    top = mkTop()
    # to_verilog() method is immuatable.
    # Hooked methods are called for the copied object to keep the internal state.
    dummy = top.to_verilog()
    verilog = top.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function

import veriloggen as vg
import simulation_simulator_verilator

expected_rslt = """\
LED:  0 count:         0
LED:  0 count:         1
LED:  0 count:         2
LED:  0 count:         3
LED:  0 count:         4
LED:  0 count:         5
LED:  0 count:         6
LED:  0 count:         7
LED:  0 count:         8
LED:  0 count:         9
LED:  0 count:        10
LED:  0 count:         0
LED:  0 count:         1
LED:  0 count:         2
LED:  0 count:         3
LED:  0 count:         4
LED:  0 count:         5
LED:  0 count:         6
LED:  0 count:         7
LED:  0 count:         8
LED:  0 count:         9
LED:  0 count:        10
LED:  0 count:        11
LED:  0 count:        12
LED:  0 count:        13
LED:  0 count:        14
LED:  0 count:        15
LED:  0 count:        16
LED:  0 count:        17
LED:  0 count:        18
LED:  0 count:        19
LED:  0 count:        20
LED:  0 count:        21
LED:  0 count:        22
LED:  0 count:        23
LED:  0 count:        24
LED:  0 count:        25
LED:  0 count:        26
LED:  0 count:        27
LED:  0 count:        28
LED:  0 count:        29
LED:  0 count:        30
LED:  0 count:        31
LED:  1 count:         0
LED:  1 count:         1
LED:  1 count:         2
LED:  1 count:         3
LED:  1 count:         4
LED:  1 count:         5
LED:  1 count:         6
LED:  1 count:         7
LED:  1 count:         8
LED:  1 count:         9
LED:  1 count:        10
LED:  1 count:        11
LED:  1 count:        12
LED:  1 count:        13
LED:  1 count:        14
LED:  1 count:        15
LED:  1 count:        16
LED:  1 count:        17
LED:  1 count:        18
LED:  1 count:        19
LED:  1 count:        20
LED:  1 count:        21
LED:  1 count:        22
LED:  1 count:        23
LED:  1 count:        24
LED:  1 count:        25
LED:  1 count:        26
LED:  1 count:        27
LED:  1 count:        28
LED:  1 count:        29
LED:  1 count:        30
LED:  1 count:        31
LED:  2 count:         0
LED:  2 count:         1
LED:  2 count:         2
LED:  2 count:         3
LED:  2 count:         4
LED:  2 count:         5
LED:  2 count:         6
LED:  2 count:         7
LED:  2 count:         8
LED:  2 count:         9
LED:  2 count:        10
LED:  2 count:        11
LED:  2 count:        12
LED:  2 count:        13
LED:  2 count:        14
LED:  2 count:        15
"""


def test():
    vg.reset()
    test_module = simulation_simulator_verilator.mkTest()
    sim = vg.simulation.Simulator(test_module, sim='verilator')
    rslt = sim.run(sim_time=1000)
    rslt = '\n'.join([line for line in rslt.splitlines() if line.startswith('LED:')] + [''])
    assert(expected_rslt == rslt)
